import logging
import sys
import os
import io
import socket
import errno
import time

from .httpurl import native_str

__all__ = ['IStream',
           'BaseSocket',
           'Socket',
           'TCPSocket',
           'TCP6Socket',
           'wrap_socket',
           'parse_address',
           'get_socket_timeout',
           'create_socket',
           'create_client_socket',
           'server_client_sockets',
           'create_connection',
           'create_socket_address',
           'socket_pair',
           'server_socket']

LOGGER = logging.getLogger('pulsar.sock')

ALLOWED_ERRORS = (errno.EAGAIN, errno.ECONNABORTED,
                  errno.EWOULDBLOCK, errno.EPIPE,
                  errno.EINVAL)

MAXFD = 1024

def parse_address(netloc, default_port=8000):
    '''Parse an address and return a tuple with host and port'''
    netloc = native_str(netloc)
    if netloc.startswith("unix:"):
        return netloc.split("unix:")[1]
    # get host
    if '[' in netloc and ']' in netloc:
        host = netloc.split(']')[0][1:].lower()
    elif ':' in netloc:
        host = netloc.split(':')[0].lower()
    elif netloc == "":
        host = "0.0.0.0"
    else:
        host = netloc.lower()
    #get port
    netloc = netloc.split(']')[-1]
    if ":" in netloc:
        port = netloc.split(':', 1)[1]
        if not port.isdigit():
            raise RuntimeError("%r is not a valid port number." % port)
        port = int(port)
    else:
        port = default_port 
    return (host, port)

def get_socket_timeout(val):
    '''Obtain a valid stocket timeout value from *val*'''
    if val is None:
        return val  # blocking socket
    else:
        val = float(val)
    if val < 0:
        return None # Negative values for blocking sockets
    else:
        ival = int(val)
        return ival if ival == val else val

def create_connection(address, blocking=0):
    sock_type, address = create_socket_address(address)
    s = sock_type(is_server=False)
    s.sock.connect(address)
    s.sock.setblocking(blocking)
    return s

def create_socket(address, logger=None, backlog=2048,
                  bound=False, retry=5, retry_lag=2,
                  retry_step=2, interval_max=30):
    """Create a new server :class:`Socket` for the given address.
If the address is a tuple, a TCP socket is created.
If it is a string, a Unix socket is created.
Otherwise a TypeError is raised.

:parameter address: Socket address.
:parameter logger: Optional python logger instance.
:parameter backlog: The maximum number of pending connections or ``None``.
    if ``None`` this is a client socket.
:parameter bound: If ``False`` the socket will bind to *address* otherwise
    it is assumed to be already bound.
:parameter retry: Number of retries before aborting.
:parameter retry_lag: Number of seconds between connection attempts.
:rtype: Instance of :class:`Socket`
    """
    sock_type, address = create_socket_address(address)
    logger = logger or LOGGER
    lag = retry_lag
    for i in range(retry):
        try:
            return sock_type(address, backlog=backlog, bound=bound)
        except socket.error as e:
            if e.errno == errno.EADDRINUSE:
                logger.error("Connection in use: %s", address)
            elif e.errno in (errno.EADDRNOTAVAIL, 11004, -5):
                raise RuntimeError("Invalid address: %s" % str(address))
            if i < retry:
                logger.error("Retrying in %s seconds.", lag)
                time.sleep(lag)
                lag = min(lag + retry_step, interval_max)
    raise RuntimeError("Can't connect to %s. Tried %s times." %
                       (str(address), retry))


create_client_socket = lambda address: create_socket(address, bound=True)

def wrap_socket(sock):
    '''Wrap a python socket with pulsar :class:`Socket`.'''
    if sock and not isinstance(sock, Socket):
        address = sock.getsockname()
        sock_type, address = create_socket_address(address)
        return sock_type(fd=sock, bound=True, backlog=None)
    else:
        return sock

def socket_pair(backlog=2048, logger=None, blocking=0):
    '''Create a ``127.0.0.1`` (client,server) socket pair on any
available port. The first socket is connected to the second, the server socket,
which is bound to ``127.0.0.1`` at any available port.

:param backlog: number of connection to listen.
:param logger: optional python logger.
:rtype: tuple with two instances of :class:`Socket`
'''
    remote_client = socket.socket()
    remote_client.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
    count = 0
    while 1:
        count += 1
        server = create_socket(('127.0.0.1',0), logger=logger, backlog=backlog)
        try:
            # Connect the remote socket with the server socket
            remote_client.connect(server.name)
            break
        except socket.error as e:
            if e[0] != errno.WSAEADDRINUSE:
                raise
            if count >= 10:
                remote_client.close()
                server.close()
                raise socket.error("Cannot bind socket pairs!")
            remote_client.close()

    remote_client = server.__class__(fd=remote_client, bound=True, backlog=None)
    remote_client.setblocking(blocking)
    return remote_client, server

def server_client_sockets(backlog=2048, blocking=0):
    '''Create a server_connection, client pair.'''
    # get a socket pair
    client, server = socket_pair(backlog=backlog)
    server.setblocking(True)
    server_connection, _ = server.accept()
    server_connection.setblocking(blocking)
    client.setblocking(blocking)
    server_connection = server.__class__(fd=server_connection,
                                         bound=True, backlog=None)
    return server_connection, client

def server_socket(backlog=2048, blocking=0):
    '''Create a TCP socket ready for accepting connections.'''
    # get a socket pair
    w, s = socket_pair(backlog=backlog)
    s.setblocking(True)
    r, _ = s.accept()
    r.close()
    w.close()
    s.setblocking(blocking)
    return s


class IStream(object):
    '''Interface for all streams.

.. attribute: address

    The address of the stream. This is usually the address of the listening
    or writing socket.'''
    address = ('0.0.0.0', 0)

    def close(self, msg=None):
        '''Close the stream'''
        pass

    def fileno(self):
        '''Return the file descriptor of the socket.'''
        pass

    def write(self, data):
        '''Write *data* into the stream'''
        raise NotImplementedError()


class BaseSocket(IStream):
    '''Base class for :class:`IStream` using a socket as I/O mechanism.
    
.. attribute:: sock
    
    The underlying socket
    
.. attribute:: async
    
    True if this is an asynchronous socket.
    
.. attribute:: address
    
    same as :meth:`getsockname`.
'''
    sock = None
    
    def getsockname(self):
        '''The socket name if open otherwise None.'''
        try:
            return self.sock.getsockname()
        except:
            return None
    
    def settimeout(self, value):
        self.sock.settimeout(value)
        
    def gettimeout(self):
        if self.sock:
            return self.sock.gettimeout()
    
    @property
    def async(self):
        return self.gettimeout() == 0
    
    @property
    def address(self):
        return self.getsockname()
    
    def fileno(self):
        if self.sock:
            return self.sock.fileno()
    
    
class Socket(BaseSocket):
    '''Wrapper class for a python socket. It provides with
higher level tools for creating and reusing sockets already created.'''
    def __init__(self, address=None, backlog=2048, fd=None, bound=False,
                 is_server=None):
        self.backlog = backlog
        self._is_server = is_server if is_server is not None else not bound
        self._init(fd, address, bound)

    def _init(self, fd, address, bound):
        if fd is None:
            sock = socket.socket(self.FAMILY, socket.SOCK_STREAM)
        elif hasattr(fd, 'fileno'):
            self.sock = fd
            return
        else:
            sock = socket.fromfd(fd, self.FAMILY, socket.SOCK_STREAM)
        if self.is_server:
            self.sock = self.set_options(sock, address, bound)
        else:
            self.sock = sock

    @property
    def closed(self):
        return self.sock == None

    def __getstate__(self):
        d = self.__dict__.copy()
        d['fd'] = d.pop('sock').fileno()
        return d

    def __setstate__(self, state):
        fd = state.pop('fd')
        self.__dict__ = state
        self._init(fd, None, True)

    @property
    def is_server(self):
        return self._is_server

    def write(self, data):
        '''Same as the socket send method but it close the connection if
not data was sent. In this case it also raises a socket error.'''
        if self.closed or not data:
            return 0
        try:
            sent = self.send(data)
            if sent == 0:
                raise socket.error()
            return sent
        except:
            self.close()
            raise

    def accept(self):
        '''Wrap the socket accept method.'''
        try:
            return self.sock.accept()
        except socket.error as e:
            if e.errno not in ALLOWED_ERRORS:
                raise
            else:
                return None, None

    def recv(self, length=None):
        return self.sock.recv(length or io.DEFAULT_BUFFER_SIZE)

    @property
    def name(self):
        return self.address

    def __str__(self, name):
        return "<socket %d>" % self.sock.fileno()

    def __repr__(self):
        s = str(self)
        v = "<socket %d>" % self.sock.fileno()
        return s if s == v else '%s %s' % (v, s)

    def __getattr__(self, name):
        return getattr(self.sock, name)

    def set_options(self, sock, address, bound):
        '''Options for a server socket'''
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        # If the socket is not bound, bind it to the address
        if not bound:
            self.bind(sock, address)
        if self.backlog:
            sock.setblocking(0)
            sock.listen(self.backlog)
        return sock

    def bind(self, sock, address):
        sock.bind(address)

    def close(self, log=None):
        '''Shutdown and close the socket.'''
        if not self.closed:
            try:
                self.sock.shutdown(socket.SHUT_RDWR)
                self.sock.close()
            except:
                pass
            self.sock = None

    def info(self):
        if self.is_server:
            return 'listening at {0}'.format(self)
        else:
            return 'client at {0}'.format(self)


class TCPSocket(Socket):
    FAMILY = socket.AF_INET

    def __str__(self):
        name = self.name
        if name:
            return "%s:%d" % name
        else:
            return '<closed>'

    def set_options(self, sock, address, bound):
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        return super(TCPSocket, self).set_options(sock, address, bound)


class TCP6Socket(TCPSocket):
    FAMILY = socket.AF_INET6

    def __str__(self):
        (host, port, fl, sc) = self.name
        return "[%s]:%d" % (host, port)


def create_tcp_socket_address(addr):
    if not isinstance(addr, tuple):
        hp = addr.split(':')
        if len(hp) == 2:
            addr = tuple(hp)
    if isinstance(addr, tuple):
        if len(addr) != 2:
            raise ValueError('TCP address must be a (host, port) tuple')
        host = addr[0]
        try:
            port = int(addr[1])
        except:
            raise ValueError('TCP address must be a (host, port) tuple')
        if not host:
            host = '0.0.0.0'
        addr = (host, port)
        if is_ipv6(host):
            sock_type = TCP6Socket
        else:
            sock_type = TCPSocket
    else:
        raise TypeError("Unable to create socket from: %r" % addr)
    return sock_type, addr

if os.name == 'posix':

    def is_ipv6(addr):
        try:
            socket.inet_pton(socket.AF_INET6, addr)
        except socket.error: # not a valid address
            return False
        return True


    class UnixSocket(Socket):

        FAMILY = socket.AF_UNIX

        def __str__(self):
            return "unix:%s" % self.address

        def bind(self, sock, address):
            try:
                os.remove(address)
            except OSError:
                pass
            #old_umask = os.umask(self.conf.umask)
            sock.bind(address)
            #system.chown(address, self.conf.uid, self.conf.gid)
            #os.umask(old_umask)

        def close(self):
            name = self.name
            super(UnixSocket, self).close()
            if name:
                try:
                    os.remove(name)
                except OSError:
                    pass


    def create_socket_address(addr):
        """Create a new socket for the given address. If the
        address is a tuple, a TCP socket is created. If it
        is a string, a Unix socket is created. Otherwise
        a TypeError is raised.
        """
        try:
            return create_tcp_socket_address(addr)
        except TypeError:
            if isinstance(addr, str):
                sock_type = UnixSocket
            else:
                raise TypeError("Unable to create socket from: %r" % addr)
        return sock_type, addr

else:   #pragma    nocover

    def is_ipv6(addr):
        return False

    create_socket_address = create_tcp_socket_address




from collections import Mapping

import pulsar
from pulsar.apps import wsgi
from pulsar.apps import rpc

from .exceptions import TaskNotAvailable
from .task import Task


__all__ = ['TaskQueueRpcMixin']


def task_to_json(task):
    if task:
        if pulsar.is_failure(task):
            err = task.trace[1]
            if isinstance(err, TaskNotAvailable):
                raise rpc.InvalidParams('Job "%s" is not available.'\
                                        % err.task_name)
        if isinstance(task, (list, tuple)):
            task = [task_to_json(t) for t in task]
        elif isinstance(task, Task):
            task = task.tojson()
    return task
    

class TaskQueueRpcMixin(rpc.JSONRPC):
    '''A :class:`pulsar.apps.rpc.JSONRPC` mixin for communicating with
a :class:`TaskQueue`.
To use it, you need to have an :ref:`RPC application <apps-rpc>`
and a :ref:`task queue <apps-tasks>` application installed in the
:class:`pulsar.Arbiter`.

:parameter taskqueue: instance or name of the
    :class:`pulsar.apps.tasks.TaskQueue` which exposes the remote procedure
    calls.
    

**Remote Procedure Calls**

.. method:: job_list([jobnames=None])

    Return the list of :class:`Job` registered with task queue with meta
    information. If a list of jobnames is given, it returns only jobs
    included in the list.
    
    :rtype: A list of dictionaries
    
    
.. method:: run_new_task(jobname, [**kwargs])
    
    Run a new :class:`Task` in the task queue. The task can be of any type
    as long as it is registered in the :class:`Job` registry.

    :parameter jobname: the name of the :class:`Job` to run.
    :parameter kwargs: optional key-valued job parameters.
    :rtype: a dictionary containing information about the
        :class:`Task` submitted
    
    
.. method:: get_task(id=task_id)

    Retrieve a task from its ``id``.
    Returns ``None`` if the task is not available.
    
    
.. method:: get_tasks(**filters)

    Retrieve a list of tasks which satisfy *filters*.
    
    
.. method:: wait_for_task(id=task_id)

    Wait for a task to have finished.
'''
    def __init__(self, taskqueue, **kwargs):
        if not isinstance(taskqueue, str):
            taskqueue = taskqueue.name
        self.taskqueue = taskqueue
        super(TaskQueueRpcMixin,self).__init__(**kwargs)
        
    ############################################################################
    ##    REMOTES
    def rpc_job_list(self, request, jobnames=None):
        return self._rq(request, 'job_list', jobnames=jobnames)
    
    def rpc_next_scheduled_tasks(self, request, jobnames=None):
        return self._rq(request, 'next_scheduled', jobnames=jobnames)
        
    def rpc_run_new_task(self, request, jobname=None, **kw):
        result = self.run_new_task(request, jobname, **kw)
        return result.addBoth(task_to_json)
        
    def rpc_get_task(self, request, id=None):
        if id:
            return self._rq(request, 'get_task', id).add_callback(task_to_json)
    
    def rpc_get_tasks(self, request, **params):
        if params:
            return self._rq(request, 'get_tasks', **params).add_callback(task_to_json)
        
    def rpc_wait_for_task(self, request, id=None):
        if id:
            return self._rq(request, 'wait_for_task', id).add_callback(task_to_json)
    
    ############################################################################
    ##    INTERNALS
    def run_new_task(self, request, jobname, args=None,
                     ack=True, meta_data=None, **kw):
        if not jobname:
            raise rpc.InvalidParams('"jobname" is not specified!')
        funcname = 'addtask' if ack else 'addtask_noack'
        meta_data = meta_data or {}
        meta_data.update(self.task_request_parameters(request))
        args = args or ()
        return self._rq(request, funcname, jobname, meta_data, *args, **kw)
    
    def task_request_parameters(self, request):
        '''**Internal function** which returns a dictionary of parameters
to be passed to the :class:`Task` class constructor.
This function can be overridden to add information about
the type of request, who made the request and so forth. It must return
a dictionary. By default it returns an empty dictionary.'''
        return {}
    
    def _rq(self, request, action, *args, **kw):
        return pulsar.send(self.taskqueue, action, *args, **kw)
    



'''\
A Python implementation of JSON-RPC 2.0 Specification

JSON-RPC is a lightweight remote procedure call protocol.
It's designed to be simple.

A remote method is invoked by sending a request to a remote service.
The request is a single object serialized using JSON.

The specification is at http://groups.google.com/group/json-rpc/web/json-rpc-2-0
'''
import logging
import json
from timeit import default_timer

import pulsar
from pulsar import HttpClient, is_async
from pulsar.utils.structures import AttributeDictionary
from pulsar.utils.security import gen_unique_id
from pulsar.utils.httpurl import to_string, range
from pulsar.utils.jsontools import DefaultJSONEncoder, DefaultJSONHook

from .handlers import RpcHandler, Request
from .exceptions import exception, INTERNAL_ERROR, REQUIRES_AUTHENTICATION

__all__ = ['JSONRPC', 'JsonProxy', 'LocalJsonProxy']


class JsonToolkit(object):

    @classmethod
    def dumps(cls, data, **kwargs):
        return json.dumps(data, cls=DefaultJSONEncoder, **kwargs)

    @classmethod
    def loads(cls, content):
        return json.loads(to_string(content), object_hook=DefaultJSONHook)


class JSONRPC(RpcHandler):
    '''A :class:`RpcHandler` for class for JSON-RPC services.
Design to comply with the `JSON-RPC 2.0`_ Specification.

.. _`JSON-RPC 2.0`: http://groups.google.com/group/json-rpc/web/json-rpc-2-0'''
    content_type = 'application/json'
    methods = ('post',)
    _json = JsonToolkit

    def get_method_and_args(self, data):
        '''Overrides the :meth:`RpcHandler:get_method_and_args` to obtain
method data from the JSON *data* string.'''
        if not isinstance(data, dict):
            data = self._json.loads(data)
        method  = data.get('method',None)
        params  = data.get('params',None)
        id      = data.get('id',None)
        version = data.get('jsonrpc',None)
        kwargs  = {}
        args    = ()
        if isinstance(params, dict):
            for k, v in params.items():
                kwargs[str(k)] = v
        elif params:
            args = tuple(params)
        return method, args, kwargs, id, version

    def dumps(self, id, version, result=None, error=None):
        '''Modify JSON dumps method to comply with JSON-RPC Specification 2.0'''
        res = {'id': id, "jsonrpc": version}
        if error:
            if hasattr(error, 'faultCode'):
                code = error.faultCode
                msg = getattr(error, 'faultString', str(error))
            else:
                msg = str(error)
                if getattr(error, 'status', None) == 403:
                    code = REQUIRES_AUTHENTICATION
                else:
                    code = INTERNAL_ERROR
            res['error'] =  {'code': code,
                             'message': msg,
                             'data': getattr(error,'data','')}
        else:
            res['result'] = result
        return self._json.dumps(res)


class JsonCall:
    
    dlots = ('_client', '_name')
    
    def __init__(self, client, name):
        self._client = client
        self._name = name
        
    def __repr__(self):
        return self._name
    __str__ = __repr__
    
    @property
    def url(self):
        return self._client.url
    
    @property
    def name(self):
        return self._name
    
    def __getattr__(self, name):
        name = "%s%s%s" % (self._name, self._client.separator, name)
        return self.__class__(self._client, name)
        
    def __call__(self, *args, **kwargs):
        c = self._client
        data, raw = c._get_data(self._name, *args, **kwargs)
        body = c._json.dumps(data).encode('latin-1')
        # Always make sure the content-type is application/json
        c.http.headers['content-type'] = 'application/json'
        resp = c.http.post(c.url, data=body)
        if self._client._full_response:
            return resp
        elif is_async(resp):
            return resp.add_callback(lambda r: self._end_call(r, raw))
        else:
            return self._end_call(resp, raw)
    
    def _end_call(self, resp, raw):
        content = resp.content.decode('utf-8')
        if resp.is_error:
            if 'error' in content:
                return self._client.loads(content)
            else:
                resp.raise_for_status()
        else:
            if raw:
                return content
            else:
                return self._client.loads(content)
        
class JsonProxy(object):
    '''A python Proxy class for :class:`JSONRPC` Servers.

:param url: server location
:param version: JSONRPC server version. Default ``2.0``
:param id: optional request id, generated if not provided. Default ``None``.
:param data: Extra data to include in all requests. Default ``None``.
:param http: optional http opener. If provided it must have the ``request``
    method available which must be of the form::

        http.request(url, body=..., method=...)

    Default ``None``.

Lets say your RPC server is running at ``http://domain.name.com/``::

    >>> a = JsonProxy('http://domain.name.com/')
    >>> a.add(3,4)
    7
    >>> a.ping()
    'pong'

'''
    separator  = '.'
    rawprefix  = 'raw'
    default_version = '2.0'
    default_timeout = 3
    _json = JsonToolkit

    def __init__(self, url, version=None, data=None, full_response=False, **kw):
        self.__url = url
        self.__version = version or self.__class__.default_version
        self._full_response = full_response
        self.__data = data if data is not None else {}
        self.local = AttributeDictionary()
        self.setup(**kw)

    def setup(self, http=None, timeout=None, **kw):
        if not http:
            timeout = timeout if timeout is not None else self.default_timeout
            http = HttpClient(timeout=timeout, **kw)
        self.local.http = http

    @property
    def url(self):
        return self.__url

    @property
    def http(self):
        return self.local.http

    def makeid(self):
        '''Can be re-implemented by your own Proxy'''
        return gen_unique_id()

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.__url)

    def __str__(self):
        return self.__repr__()
    
    def __getattr__(self, name):
        return JsonCall(self, name)

    def timeit(self, func, times, *args, **kwargs):
        '''Usefull little utility for timing responses from server. The
usage is simple::

    >>> from pulsar.apps import rpc
    >>> p = rpc.JsonProxy('http://127.0.0.1:8060')
    >>> p.timeit('ping',10)
    0.56...
    >>> _
    '''
        r = range(times)
        func = getattr(self, func)
        start = default_timer()
        for t in r:
            func(*args, **kwargs)
        return default_timer() - start

    def _get_data(self, func_name, *args, **kwargs):
        id = self.makeid()
        fs = func_name.split('_')
        raw = False
        if len(fs) > 1 and fs[0] == self.rawprefix:
            raw = True
            fs.pop(0)
            func_name = '_'.join(fs)
        params = self.get_params(*args, **kwargs)
        data = {'method': func_name, 'params': params, 'id': id,
                'jsonrpc': self.__version}
        return data, raw

    def get_params(self, *args, **kwargs):
        '''
        Create an array or positional or named parameters
        Mixing positional and named parameters in one
        call is not possible.
        '''
        kwargs.update(self.__data)
        if args and kwargs:
            raise ValueError('Cannot mix positional and named parameters')
        if args:
            return list(args)
        else:
            return kwargs

    def loads(self, obj):
        res = self._json.loads(obj)
        if isinstance(res, dict):
            if 'error' in res:
                error = res['error']
                code    = error['code']
                message = error['message']
                raise exception(code, message)
            else:
                return res.get('result',None)
        return res


class LocalJsonProxy(JsonProxy):
    '''A proxy class to use when accessing the rpc within the rpc application
domain.'''
    def setup(self, handler=None, environ=None, **kwargs):
        self.local.handler = handler
        self.local.environ = environ

    def __call__(self, *args, **kwargs):
        data, raw = self._get_data(*args, **kwargs)
        hnd = self.local.handler
        environ = self.local.environ
        method, args, kwargs, id, version = hnd.get_method_and_args(data)
        hnd.request(environ, method, args, kwargs, id, version)
        return environ['rpc'].process(Request(environ))



__all__ = ['Request']

class Request(object):
    '''Wsgi environment wrapper'''

    def __init__(self, environ):
        self.environ = environ

    def __repr__(self):
        return self.environ.__repr__()
    
    def __str__(self):
        return self.__repr__()

    def __getattr__(self, name):
        if name in self.environ:
            return self.environ[name]
        elif name in self.__dict__:
            return self.__dict__[name]
        else:
            raise AttributeError("'%s' object has no attribute '%s'" %\
                                 (self.__class__.__name__, name))

    def __getitem__(self, name):
        return self.environ[name]
    
    def __setitem__(self, name, value):
        self.environ[name] = value
        
    def __len__(self):
        return len(self.environ)
    
    def __iter__(self):
        return iter(self.environ)
    


'''\
A a JSON-RPC Server with some simple functions.
To run the server type::

    python manage.py

Open a new shell and launch python and type::

    >>> from pulsar.apps import rpc
    >>> p = rpc.JsonProxy('http://localhost:8060')
    >>> p.ping()
    'pong'
    >>> p.functions_list()
    >>> ...
    >>> p.calc.add(3,4)
    7.0
    >>>

'''
try:
    import pulsar
except ImportError: #pragma nocover
    import sys
    sys.path.append('../../')
from random import normalvariate

from pulsar.apps import rpc, wsgi
from pulsar.utils.httpurl import range


def divide(request, a, b):
    '''Divide two numbers'''
    return float(a)/float(b)

def request_handler(request, format, kwargs):
    '''Dummy request handler'''
    return kwargs

def randompaths(request, num_paths=1, size=250, mu=0, sigma=1):
    r = []
    for p in range(num_paths):
        v = 0
        path = [v]
        r.append(path)
        for t in range(size):
            v += normalvariate(mu, sigma)
            path.append(v)
    return r


class RequestCheck:
    
    def __call__(self, request, name):
        method = request.rpc.method
        assert(method==name)
        return True


class Root(rpc.PulsarServerCommands):
    
    def rpc_dodgy_method(self, request):
        '''This method will fails because the return object is not
json serializable.'''
        return Calculator
    
    rpc_check_request = RequestCheck()


class Calculator(rpc.JSONRPC):

    def rpc_add(self, request, a, b):
        return float(a) + float(b)

    def rpc_subtract(self, request, a, b):
        return float(a) - float(b)

    def rpc_multiply(self, request, a, b):
        return float(a) * float(b)

    rpc_divide = rpc.FromApi(divide, request_handler=request_handler)
    rpc_randompaths = rpc.FromApi(randompaths)


def wsgi_handler():
    return rpc.RpcMiddleware(Root().putSubHandler('calc',Calculator()))

def server(callable=None, **params):
    return wsgi.WSGIServer(callable=wsgi_handler(), **params)


if __name__ == '__main__':  #pragma nocover
    server().start()




'''\
A task scheduler application with HTTP-RPC hooks
'''
from datetime import datetime
import logging
import traceback
from io import StringIO

from pulsar.utils.httpurl import itervalues, iteritems
from pulsar import maybe_async, as_failure, is_async, is_failure, send

from .models import registry
from .exceptions import *
from .states import *


__all__ = ['Task', 'TaskInMemory', 'TaskConsumer', 'nice_task_message']


class TaskConsumer(object):
    '''A context manager for consuming tasks. This is used by the
:class:`Scheduler` in a ``with`` block to correctly handle exceptions
and optional logging.

.. attribute:: task

    the :class:`Task` being consumed.

.. attribute:: job

    the :class:`Job` which generated the :attr:`task`.

.. attribute:: worker

    the :class:`pulsar.apps.Worker` running the process.
'''
    def __init__(self, task, worker, job):
        self.worker = worker
        self.job = job
        self.task = task


class Task(object):
    '''Interface for tasks which are produced by :class:`Job`.

.. attribute:: id

    :class:`Task` unique id.

.. attribute:: name

    :class:`Job` name.

.. attribute:: status

    The current :ref:`status string <task-state>` of task.

.. attribute:: time_executed

    date time when the task was executed.

.. attribute:: time_start

    date-time when the task calculation has started.

.. attribute:: time_end

    date-time when the task has finished.

.. attribute:: expiry

    optional date-time indicating when the task should expire.

.. attribute:: timeout

    A datetime or ``None`` indicating whether a timeout has occurred.

.. attribute:: from_task

    Optional :attr:`Task.id` for the :class:`Task` which queued
    this :class:`Task`. This is a usuful for monitoring the creation
    of tasks within other tasks.
'''
    from_task = None
    stack_trace = None

    @property
    def status_code(self):
        '''Integer indicating :attr:`status` precedence.
Lower number higher precedence.'''
        return PRECEDENCE_MAPPING.get(self.status, UNKNOWN_STATE)

    def start(self, worker):
        '''Called by the :class:`pulsar.Worker` *worker* when the task
start its execution. If no timeout has occured the task will switch to
a ``STARTED`` :attr:`Task.status` and invoke the :meth:`on_start`
callback.'''
        job = registry[self.name]
        result = None
        try:
            if self.maybe_revoked():
                yield self.on_timeout(worker)
            else:
                self.status = STARTED
                self.time_start = datetime.now()
                yield self.on_start(worker)
                consumer = TaskConsumer(self, worker, job)
                result = maybe_async(job(consumer, *self.args, **self.kwargs))
                if is_async(result):
                    yield result
                    result = maybe_async(result)
        except Exception as e:
            result = as_failure(e)
        finally:
            yield self.finish(worker, result)

    def finish(self, worker, result):
        '''Called when the task has finished and a ``result`` is ready.
It sets the :attr:`time_end` attribute if not already set
(in case the :class:`Task` was revoked) and
determined if it was succesful or a failure. Once done it invokes
the :ref:`task callback <tasks-callbacks>` :meth:`on_finish`.'''
        if not self.time_end:
            self.time_end = datetime.now()
            if is_failure(result):
                result.log()
                exception = result.trace[1]
                self.status = getattr(exception, 'status', FAILURE)
                self.result = str(result) if result else str(exception)
            # If the status is STARTED this is a succesful task
            elif self.status == STARTED:
                self.status = SUCCESS
                self.result = result
        return self.on_finish(worker)

    def to_queue(self, schedulter=None):
        '''The task has been received by the scheduler. If its status
is PENDING swicth to RECEIVED, save the task and return it. Otherwise
returns nothing.'''
        self._toqueue = False
        if self.status == PENDING:
            self.status = RECEIVED
            self._toqueue = True
            self.on_received(schedulter)
        return self

    def needs_queuing(self):
        '''called after calling :meth:`to_queue`, it return ``True`` if the
task needs to be queued.'''
        return self.__dict__.pop('_toqueue', False)

    def done(self):
        '''Return ``True`` if the :class:`Task` has finshed
(its status is one of :ref:`READY_STATES <task-state>`).'''
        return self.status in READY_STATES

    def maybe_revoked(self):
        '''Try to revoke a task. It returns the :attr:`timeout` which is
different from ``None`` only when the :class:`Task` has been revoked.

:rtype: ``None`` or a DateTime.'''
        if not self.timeout:
            if self.status_code > PRECEDENCE_MAPPING[STARTED] and self.expiry:
                tm = datetime.now()
                if tm > self.expiry:
                    self.status = REVOKED
                    self.timeout = tm
        return self.timeout

    def execute2start(self):
        if self.time_start:
            return self.time_start - self.time_executed

    def execute2end(self):
        if self.time_end:
            return self.time_end - self.time_executed

    def duration(self):
        '''The :class:`Task` duration. Only available if the task status is in
:attr:`FULL_RUN_STATES`.'''
        if self.time_end and self.time_start:
            return self.time_end - self.time_start

    def tojson(self):
        '''Convert the task instance into a JSON-serializable dictionary.'''
        return self.__dict__.copy()

    def serialize_for_queue(self):
        return self

    def ack(self):
        return self

    ############################################################################
    # CALLBACKS
    ############################################################################

    def on_created(self, scheduler=None):
        '''A :ref:`task callback <tasks-callbacks>` when the task has
has been created.

:parameter scheduler: the scheduler which created the task.
'''
        pass

    def on_received(self, scheduler=None):
        '''A :ref:`task callback <tasks-callbacks>` when the task has
has been received by the scheduler.'''
        pass

    def on_start(self, worker=None):
        '''A :ref:`task callback <tasks-callbacks>` when the task starts
its execution'''
        pass

    def on_timeout(self, worker=None):
        '''A :ref:`task callback <tasks-callbacks>` when the task is expired'''
        pass

    def on_finish(self, worker=None):
        '''A :ref:`task callback <tasks-callbacks>` when the task finish
its execution'''
        pass

    def close(self):
        pass

    def emit_log(self, record):
        '''Implement the task logging emit method. By default it does nothing.
It can be reimplemented to do something with the log record.'''
        pass
    
    ############################################################################
    ##    FACTORY METHODS
    ############################################################################
    @classmethod
    def get_task(cls, scheduler, id):
        '''Given a task *id* it retrieves a task instance or ``None`` if
not available.'''
        raise NotImplementedError()
    
    @classmethod
    def get_tasks(cls, scheduler, **filters):
        '''Given *filters* it retrieves task instances which satisfy the
filter criteria.'''
        raise NotImplementedError()
    
    @classmethod
    def save_task(cls, scheduler, task):
        raise NotImplementedError()
        
    @classmethod
    def delete_tasks(cls, scheduler, task):
        raise NotImplementedError()
    
    @classmethod
    def wait_for_task(cls, scheduler, id, timeout):
        '''Wait for a task to have finished.'''
        raise NotImplementedError()


class TaskInMemory(Task):
    '''An in memory implementation of a Task.'''
    time_start = None
    time_end = None
    stack_trace = None
    result = None
    timeout = False

    def __init__(self, id=None, name=None, time_executed=None,
                 expiry=None, args=None, kwargs=None, ack=None,
                 status=None, from_task=None, **params):
        self.id = id
        self.name = name
        self.time_executed = time_executed
        self.expiry = expiry
        self.args = args
        self.kwargs = kwargs
        self.status = status
        self.from_task = from_task
        self.params = params

    def __repr__(self):
        return '%s(%s)' % (self.name,self.id)
    __str__ = __repr__

    def on_received(self, scheduler=None):
        # Called by the scheduler
        self.save_task(scheduler, self)

    def on_start(self, worker=None):
        if worker:
            return worker.send(worker.monitor, 'save_task', self)

    def on_timeout(self, worker=None):
        if worker:
            return worker.send(worker.monitor, 'save_task', self)

    def on_finish(self, worker=None):
        if worker:
            return worker.send(worker.monitor, 'save_task', self)

    @classmethod
    def task_container(cls, scheduler):
        if not hasattr(scheduler, '_TASKS'):
            scheduler._TASKS = {}
        return scheduler._TASKS
    
    @classmethod
    def save_task(cls, scheduler, task):
        TASKS = cls.task_container(scheduler)
        if task.id in TASKS:
            t = TASKS[task.id]
            if t.status_code < task.status_code:
                # we don't save here. Could by concurrency lags
                return
        TASKS[task.id] = task

    @classmethod
    def delete_tasks(cls, scheduler, ids=None):
        TASKS = cls.task_container(scheduler)
        if ids:
            for id in ids:
                TASKS.pop(id, None)
        else:
            TASKS.clear()
        
    @classmethod
    def get_task(cls, scheduler, id):
        return cls.task_container(scheduler).get(id)
    
    @classmethod
    def get_tasks(cls, scheduler, **filters):
        TASKS = cls.task_container(scheduler)
        tasks = []
        if filters:
            fs = []
            for name, value in iteritems(filters):
                if not isinstance(value, (list, tuple)):
                    value = (value,)
                fs.append((name, value))
            for t in itervalues(TASKS):
                for name, values in fs:
                    value = getattr(t, name, None)
                    if value in values:
                        tasks.append(t)
        return tasks


def format_time(dt):
    return dt.isoformat() if dt else '?'

def nice_task_message(req, smart_time=None):
    smart_time = smart_time or format_time
    status = req['status'].lower()
    user = req.get('user')
    ti = req.get('time_start', req.get('time_executed'))
    name = '%s (%s) ' % (req['name'], req['id'][:8])
    msg = '%s %s at %s' % (name, status, smart_time(ti))
    return '%s by %s' % (msg, user) if user else msg



"""Multi-consumer multi-producer dispatching mechanism

Originally based on pydispatch (BSD) http://pypi.python.org/pypi/PyDispatcher/2.0.1
See license.txt for original license.
"""
from .dispatcher import Signal

_events = {}

def create(name, providing_args=None):
    '''Create a new event from a *name*. If the event is already available
it returns it.'''
    name = name.lower()
    if name not in _events:
        event = Signal()
        _events[name] = event
    return _events[name]

def bind(name, callback, sender=None, **kwargs):
    '''Bind a *callback* to event *name*. The optional *sender* can be used
to bind only when the given sender fires the event.'''
    event = create(name)
    event.connect(callback, sender=sender, **kwargs)
    
def fire(name, sender, **kwargs):
    '''Fire event *name* from *sender*.'''
    event = create(name)
    event.send(sender, **kwargs)


#First try local
try:
    from ._pulsar import *
    hasextensions = True
except ImportError: #pragma    nocover
    # Try Global
    try:
        from _pulsar import *
        hasextensions = True
    except ImportError:
        hasextensions = False
        from pulsar.utils.httpurl import HttpParser

Http_Parser = HttpParser

def setDefaultHttpParser(parser):
    global Http_Parser
    Http_Parser = parser


import io
import sys
import logging
import socket
import time
import threading
from multiprocessing.queues import Empty, Queue

from pulsar import create_connection, MailboxError, server_socket,\
                    wrap_socket, CouldNotParse, CommandNotFound,\
                    defaults
from pulsar.utils.httpurl import to_bytes

from .defer import maybe_async, pickle, is_async, log_failure,\
                    async, is_failure, ispy3k, raise_failure, CLEAR_ERRORS
from .iostream import AsyncIOStream, AsyncSocketServer,\
                        AsyncConnection, ReconnectingClient, AsyncResponse
from .access import get_actor


__all__ = ['PulsarClient', 'mailbox', 'Mailbox', 'IOQueue',
           'Empty', 'Queue', 'ActorMessage']


LOGGER = logging.getLogger('pulsar.mailbox')


def mailbox(actor=None, address=None):
    '''Creates a :class:`Mailbox` instances for :class:`Actor` instances.
If an address is provided, the communication is implemented using a socket,
otherwise a queue is used.'''
    if address:
        return PulsarClient.connect(address, timeout=0)
    else:
        if actor.is_monitor():
            return MonitorMailbox(actor)
        else:
            return Mailbox.make(actor)

def actorid(actor):
    return actor.aid if hasattr(actor, 'aid') else actor


class MessageParser(object):

    def encode(self, msg):
        if isinstance(msg, ActorMessage):
            return msg.encode()
        else:
            return to_bytes(msg)

    def decode(self, buffer):
        return ActorMessage.decode(buffer)


class ActorMessage(object):
    '''A message which travels from :class:`Actor` to
:class:`Actor` to perform a specific *command*. :class:`ActorMessage`
are not directly initialised using the constructor, instead they are
created by :meth:`ActorProxy.send` method.

.. attribute:: sender

    id of the actor sending the message.

.. attribute:: receiver

    id of the actor receiving the message.

.. attribute:: command

    command to be performed

.. attribute:: args

    Positional arguments in the message body

.. attribute:: kwargs

    Optional arguments in the message body
'''
    def __init__(self, command, sender=None, receiver=None,
                 args=None, kwargs=None):
        self.command = command
        self.sender = actorid(sender)
        self.receiver = actorid(receiver)
        self.args = args if args is not None else ()
        self.kwargs = kwargs if kwargs is not None else {}

    @classmethod
    def decode(cls, buffer):
        separator = b'\r\n'
        if buffer[0] != 42:
            raise CouldNotParse()
        idx = buffer.find(separator)
        if idx < 0:
            return None, buffer
        length = int(buffer[1:idx])
        idx += len(separator)
        total_length = idx + length
        if len(buffer) >= total_length:
            data, buffer = buffer[idx:total_length:], buffer[total_length:]
            if not ispy3k:
                data = bytes(data)
            args = pickle.loads(data)
            return cls(*args), buffer
        else:
            return None, buffer

    def encode(self):
        data = (self.command, self.sender, self.receiver,
                self.args, self.kwargs)
        bdata = pickle.dumps(data, protocol=2)
        return ('*%s\r\n' % len(bdata)).encode('utf-8') + bdata

    def __repr__(self):
        return self.command


class PulsarClient(ReconnectingClient):
    '''A proxy for the :attr:`Actor.inbox` attribute. It is used by the
:class:`ActorProxy` to send messages to the remote actor.'''
    protocol_factory = MessageParser

    def parsedata(self, data):
        msg = super(PulsarClient, self).parsedata(data)
        if msg:
            # Those two messages are special
            if msg.command in ('callback', 'errback'):
                return msg.args[0] or ''
            else:
                return msg

    @raise_failure
    def ping(self):
        return self.execute(ActorMessage('ping'))

    @raise_failure
    def echo(self, message):
        return self.execute(ActorMessage('echo', args=(message,)))

    @raise_failure
    def run_code(self, code):
        return self.execute(ActorMessage('run_code', args=(code,)))

    @raise_failure
    def info(self):
        return self.execute(ActorMessage('info'))

    @raise_failure
    def quit(self):
        return self.execute(ActorMessage('quit'))

    @raise_failure
    def shutdown(self):
        return self.execute(ActorMessage('stop'))


class MailboxResponse(AsyncResponse):
    __slots__ = ('connection', 'parsed_data')
    def __iter__(self):
        # The receiver could be different from the mail box actor. For
        # example a monitor uses the same mailbox as the arbiter
        message = self.parsed_data
        actor = self.connection.actor
        receiver = actor.get_actor(message.receiver) or actor
        sender = receiver.get_actor(message.sender)
        command = receiver.command(message.command)
        try:
            if not command:
                raise CommandNotFound(message.command)
            args = message.args
            # If this is an internal command add the sender information
            if command.internal:
                args = (sender,) + args
            result = command(self, receiver, *args, **message.kwargs)
        except:
            result = sys.exc_info()
        result = maybe_async(result)
        while is_async(result):
            yield b''
            result = maybe_async(result)
        log_failure(result)
        if command.ack:
            # Send back the result as an ActorMessage
            if is_failure(result):
                m = ActorMessage('errback', sender=receiver, args=(result,))
            else:
                m = ActorMessage('callback', sender=receiver, args=(result,))
            yield self.protocol.encode(m)


class MailboxConnection(AsyncConnection):
    '''A :class:`MailboxClient` is a socket which receives messages
from a remote :class:`Actor`.
An instance of this class is created when a new connection is made
with a :class:`Mailbox`.'''
    authenticated = False


class Mailbox(AsyncSocketServer):
    '''Mailbox for an :class:`Actor`. If the actor is a
:ref:`CPU bound worker <cpubound>`, the class:`Mailbox`
creates its own :class:`IOLoop` which runs on a separate thread
of execution.'''
    protocol_factory = MessageParser
    connection_class = MailboxConnection
    response_class = MailboxResponse

    @classmethod
    def make(cls, actor, backlog=64):
        server = super(Mailbox, cls).make(actor=actor, backlog=backlog,
                                          onthread=actor.cpubound,
                                          timeout=None)
        if not server.actor.is_arbiter():
            server.ioloop.add_callback(server.send_mailbox_address)
        return server

    def send_mailbox_address(self):
        actor = self.actor
        return actor.send(actor.monitor, 'mailbox_address', self.address)\
                    .add_callback(actor.link_actor)


class MonitorMailbox(object):
    '''A :class:`Mailbox` for a :class:`Monitor`. This is a proxy for the
arbiter inbox.'''
    active_connections = 0
    def __init__(self, actor):
        self.mailbox = actor.arbiter.mailbox

    @property
    def address(self):
        return self.mailbox.address

    @property
    def ioloop(self):
        return self.mailbox.ioloop

    def _run(self):
        pass
    
    def close(self):
        pass


class QueueWaker(object):
    '''A waker for :class:`IOQueue`. Used by CPU-bound actors.'''
    def __init__(self, queue):
        self._queue = queue
        self._fd = 'waker'

    def __str__(self):
        return '%s %s' % (self.__class__.__name__, self._fd)

    def fileno(self):
        return self._fd

    def wake(self):
        try:
            self._queue.put((self._fd, None))
        except (IOError,TypeError):
            pass

    def consume(self):
        pass

    def close(self):
        pass


class IOQueue(object):
    '''Epoll like class for a IO based on queues rather than sockets.
The interface is the same as the python epoll_ implementation.

.. _epoll: http://docs.python.org/library/select.html#epoll-objects'''
    cpubound = True
    def __init__(self, queue, actor=None):
        self._queue = queue
        self._actor = actor
        self._fds = set()
        self._empty = ()

    @property
    def queue(self):
        '''The underlying distributed queue used for I/O.'''
        return self._queue

    def register(self, fd, events=None):
        '''Register a fd descriptor with the io queue object'''
        self._fds.add(fd)

    def modify(self, fd, events=None):
        '''Modify a registered file descriptor'''
        self.unregister(fd)
        self.register(fd, events)

    def unregister(self, fd):
        '''Remove a registered file descriptor from the ioqueue object.. '''
        self._fds.discard(fd)

    def poll(self, timeout=0.5):
        '''Wait for events. timeout in seconds (float)'''
        if self._actor:
            if not self._actor.can_poll():
                return self._empty
        try:
            event = self._queue.get(timeout=timeout)
        except (Empty,IOError,TypeError,EOFError):
            return self._empty
        return (event,)

    def waker(self):
        return QueueWaker(self._queue)




import os
import sys
import logging
import traceback
import time
import signal
import errno
import bisect
import socket
from threading import current_thread

from pulsar import HaltServer, Timeout
from pulsar.utils.system import IObase, IOpoll, close_on_exec, platform, Waker
from pulsar.utils.security import gen_unique_id
from pulsar.utils.log import Synchronized
from pulsar.utils.structures import WeakList

from .defer import Deferred, is_async, maybe_async, thread_loop, make_async,\
                    log_failure, EXIT_EXCEPTIONS

__all__ = ['IOLoop', 'PeriodicCallback', 'loop_timeout']

LOGGER = logging.getLogger('pulsar.eventloop')

def file_descriptor(fd):
    if hasattr(fd, 'fileno'):
        return fd.fileno()
    else:
        return fd

def setid(self):
    self.tid = current_thread().ident
    self.pid = os.getpid()


class LoopGuard(object):
    '''Context manager for the eventloop'''
    def __init__(self, loop):
        self.loop = loop

    def __enter__(self):
        loop = self.loop
        loop.logger.subdebug("Starting event loop")
        loop._running = True
        if not loop._started:
            loop._started = time.time()
        setid(loop)
        loop._on_exit = Deferred(description='IOloop.on_exit')
        return self

    def __exit__(self, type, value, traceback):
        loop = self.loop
        loop._running = False
        loop.logger.subdebug('Exiting event loop')
        loop._on_exit.callback(loop)


class IOLoop(IObase, Synchronized):
    """\
A level-triggered I/O event loop adapted from tornado.

:parameter io: The I/O implementation. If not supplied, the best possible
    implementation available will be used. On posix system this is ``epoll``,
    or else ``select``. It can be any other custom implementation as long as
    it has an ``epoll`` like interface. Pulsar ships with an additional
    I/O implementation based on distributed queue :class:`IOQueue`.

**ATTRIBUTES**

.. attribute:: _impl

    The IO implementation

.. attribute:: cpubound

    If ``True`` this is a CPU bound event loop, otherwise it is an I/O
    event loop. CPU bound loops can block the loop for considerable amount
    of time.
        
.. attribute:: num_loops

    Total number of loops

.. attribute:: poll_timeout

    The timeout in seconds when polling with epol or select.

    Default: `0.5`

.. attribute:: tid

    The thread id where the eventloop is running
    
.. attribute:: tasks

    A list of callables to be executed at each iteration of the event loop.
    Task can be added and deleted via the :meth:`add_task` and
    :meth:`remove_task`. Extra care must be taken when adding tasks to
    I/O event loops. These tasks should be fast to perform and not block.

**METHODS**
"""
    # Never use an infinite timeout here - it can stall epoll
    poll_timeout = 0.5

    def __init__(self, io=None, logger=None, poll_timeout=None):
        self._impl = io or IOpoll()
        self.poll_timeout = poll_timeout if poll_timeout else self.poll_timeout
        self.logger = logger or LOGGER
        if hasattr(self._impl, 'fileno'):
            close_on_exec(self._impl.fileno())
        self._handlers = {}
        self._events = {}
        self._callbacks = []
        self._timeouts = []
        self._started = None
        self._running = False
        self.num_loops = 0
        self._waker = getattr(self._impl, 'waker', Waker)()
        self._on_exit = None
        self.add_handler(self._waker,
                         lambda fd, events: self._waker.consume(),
                         self.READ)

    @property
    def cpubound(self):
        return getattr(self._impl, 'cpubound', False)

    def add_handler(self, fd, handler, events):
        """Registers the given *handler* to receive the given events for the
file descriptor *fd*.

:parameter fd: A file descriptor or an object with the ``fileno`` method.
:parameter handler: A callable which will be called when events occur on the
    file descriptor *fd*.
:rtype: ``True`` if the handler was succesfully added."""
        if fd is not None:
            fdd = file_descriptor(fd)
            if fdd not in self._handlers:
                self._handlers[fdd] = handler
                self._impl.register(fdd, events | self.ERROR)
                return True
            else:
                self.logger.debug('Handler for %s already available.', fd)

    def update_handler(self, fd, events):
        """Changes the events we listen for fd."""
        self._impl.modify(file_descriptor(fd), events | self.ERROR)

    def remove_handler(self, fd):
        """Stop listening for events on fd."""
        fdd = file_descriptor(fd)
        self._handlers.pop(fdd, None)
        self._events.pop(fdd, None)
        try:
            self._impl.unregister(fdd)
        except (OSError, IOError):
            self.logger.error("Error removing %s from IOLoop", fd, exc_info=True)

    def start(self):
        if self._running:
            return False
        self._run()

    def stop(self):
        '''Stop the loop after the current event loop iteration is complete.
If the event loop is not currently running, the next call to :meth:`start`
will return immediately.

To use asynchronous methods from otherwise-synchronous code (such as
unit tests), you can start and stop the event loop like this::

    ioloop = IOLoop()
    async_method(ioloop=ioloop, callback=ioloop.stop)
    ioloop.start()

:meth:`start` will return after async_method has run its callback,
whether that callback was invoked before or after ioloop.start.'''
        if self.running():
            self._running = False
            self.wake()
        return self._on_exit

    def running(self):
        """Returns true if this IOLoop is currently running."""
        return self._running

    def add_timeout(self, deadline, callback):
        """Add a timeout *callback*. A timeout callback  it is called
at the time *deadline* from the :class:`IOLoop`.
It returns an handle that may be passed to remove_timeout to cancel."""
        timeout = _Timeout(deadline, callback)
        bisect.insort(self._timeouts, timeout)
        return timeout

    def remove_timeout(self, timeout):
        """Cancels a pending *timeout*. The argument is an handle as returned
by the :meth:`add_timeout` method."""
        self._timeouts.remove(timeout)

    def add_callback(self, callback, wake=True):
        """Calls the given callback on the next I/O loop iteration.

        It is safe to call this method from any thread at any time.
        Note that this is the *only* method in IOLoop that makes this
        guarantee; all other interaction with the IOLoop must be done
        from that IOLoop's thread.  add_callback() may be used to transfer
        control from other threads to the IOLoop's thread.
        """
        self._callbacks.append(callback)
        if wake:
            self.wake()

    def add_periodic(self, callback, period):
        """Add a :class:`PeriodicCallback` to the event loop."""
        p = PeriodicCallback(callback, period, self)
        p.start()
        return p
        
    def wake(self):
        '''Wake up the eventloop.'''
        if self.running():
            self._waker.wake()

    ############################################################ INTERNALS    
    def _run_callback(self, callback, name='callback'):
        try:
            callback()
        except EXIT_EXCEPTIONS:
            raise
        except:
            self.logger.critical('Unhandled exception in %s.', name,
                                 exc_info=True)

    def _run(self):
        """Runs the I/O loop until one of the I/O handlers calls stop(), which
will make the loop stop after the current event iteration completes."""
        with LoopGuard(self) as guard:
            while self._running:
                poll_timeout = self.poll_timeout
                self.num_loops += 1
                _run_callback = self._run_callback
                # Prevent IO event starvation by delaying new callbacks
                # to the next iteration of the event loop.
                callbacks = self._callbacks
                if callbacks:
                    self._callbacks = []
                    for callback in callbacks:
                        _run_callback(callback)
                if self._timeouts:
                    now = time.time()
                    while self._timeouts and self._timeouts[0].deadline <= now:
                        timeout = self._timeouts.pop(0)
                        self._run_callback(timeout.callback)
                    if self._timeouts:
                        milliseconds = self._timeouts[0].deadline - now
                        poll_timeout = min(milliseconds, poll_timeout)
                # A chance to exit
                if not self._running:
                    break
                try:
                    event_pairs = self._impl.poll(poll_timeout)
                except Exception as e:
                    # Depending on python version and IOLoop implementation,
                    # different exception types may be thrown and there are
                    # two ways EINTR might be signaled:
                    # * e.errno == errno.EINTR
                    # * e.args is like (errno.EINTR, 'Interrupted system call')
                    eno = getattr(e, 'errno', None)
                    if eno != errno.EINTR:
                        args = getattr(e, 'args', None)
                        if isinstance(args, tuple) and len(args) == 2:
                            eno = args[0]
                    if eno != errno.EINTR and self._running:
                        raise
                    continue
                # Pop one fd at a time from the set of pending fds and run
                # its handler. Since that handler may perform actions on
                # other file descriptors, there may be reentrant calls to
                # this IOLoop that update self._events
                if event_pairs:
                    self._events.update(event_pairs)
                    _events = self._events
                    while _events:
                        fd, events = _events.popitem()
                        try:
                            self._handlers[fd](fd, events)
                        except EXIT_EXCEPTIONS:
                            raise
                        except (OSError, IOError) as e:
                            if e.args[0] == errno.EPIPE:
                                # Happens when the client closes the connection
                                pass
                            else:
                                self.logger.error(
                                    "Exception in I/O handler for fd %s",
                                              fd, exc_info=True)
                        except KeyError:
                            self.logger.info("File descriptor %s missing", fd)
                        except:
                            self.logger.error("Exception in I/O handler for fd %s",
                                          fd, exc_info=True)


class _Timeout(object):
    """An IOLoop timeout, a UNIX timestamp and a callback"""
    __slots__ = ('deadline', 'callback')

    def __init__(self, deadline, callback):
        self.deadline = deadline
        self.callback = callback

    def __lt__(self, other):
        return ((self.deadline, id(self.callback)) <
                (other.deadline, id(other.callback)))


class PeriodicCallback(object):
    """Schedules the given callback to be called periodically.

    The callback is called every callback_time seconds.
    """
    def __init__(self, callback, callback_time, ioloop):
        self.callback = callback
        self.callback_time = callback_time
        self.ioloop = ioloop
        self._running = False

    def start(self):
        self._running = True
        timeout = time.time() + self.callback_time
        self.ioloop.add_timeout(timeout, self._run)

    def stop(self):
        self._running = False

    def _run(self):
        if not self._running:
            return
        try:
            self.callback(self)
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            LOGGER.error("Error in periodic callback", exc_info=True)
        if self._running:
            self.start()


class _not_called_exception:

    def __init__(self, value):
        self.value = value

    def __call__(self):
        if not self.value.called:
            try:
                raise Timeout('"%s" timed out.' % self.value)
            except:
                self.value.callback(sys.exc_info())


def loop_timeout(value, timeout, ioloop=None):
    value = maybe_async(value)
    if timeout and is_async(value):
        ioloop = ioloop or thread_loop()
        return ioloop.add_timeout(time.time() + timeout,
                                  _not_called_exception(value))




import os
import sys
import glob

try:
    from importlib import *
except ImportError:
    from pulsar.utils.fallbacks._importlib import *


def expand_star(mod_name):
    """Expand something like 'unuk.tasks.*' into a list of all the modules
    there.
    """
    expanded = []
    mod_dir  = os.path.dirname(__import__(mod_name[:-2], {}, {}, ['']).__file__)
    for f in glob.glob1(mod_dir, "[!_]*.py"):
        expanded.append('%s.%s' % (mod_name[:-2], f[:-3]))
    return expanded


def import_modules(modules):
    '''Safely import a list of *modules*
    '''
    for mname in modules:
        if mname.endswith('.*'):
            to_load = expand_star(mname)
        else:
            to_load = [mname]
        for module in to_load:
            try:
                __import__(module)
            except ImportError as e:
                pass
            
def module_attribute(dotpath, default=None, safe=False):
    '''
    Load an attribute from a module.
    If the module or the attriubute is not available,
    return the default argument
    '''
    if dotpath:
        bits = str(dotpath).split('.')
        try:
            module = import_module('.'.join(bits[:-1]))
            return getattr(module,bits[-1],default)
        except Exception as e:
            if not safe:
                raise
            return default
    else:
        if not safe:
            raise ImportError()
        return default

def py_file(name):
    if name.endswith('.py'):
        return name[:-3]
    elif name.endswith('.pyc'):
        return name[:-4]
    else:
        return name
    
def import_system_file(mod, add_to_path=True):
    if os.path.isfile(mod):
        # it is a file in the system path
        dir, name = os.path.split(mod)
        names = [py_file(name)]
        while dir and not dir in sys.path:
            ndir, name = os.path.split(dir)
            if dir == ndir:
                dir = ''
                break
            dir = ndir
            names.insert(0, name)
        # the file was not in the system path
        if not dir and add_to_path:
            dir, name = os.path.split(mod)
            if dir and dir != mod:
                sys.path.append(dir)
            mod_name = py_file(name)
        else:
            mod_name = '.'.join(names)
        return import_module(mod_name)



'''\
An asynchronous parallel testing suite :class:`pulsar.Application`.
Used for testing pulsar itself but it can be used as a test suite for
any other library.

Requirements
====================
* unittest2_ for python 2.6
* mock_ for python < 3.3

.. _apps-test-intro:

Introduction
====================

Create a script on the top level directory of your library,
let's call it ``runtests.py``::

    from pulsar.apps import TestSuite

    if __name__ == '__main__':
        TestSuite(description='Test suite for my library',
                  modules=('regression',
                           ('examples','tests'))).start()

where ``modules`` is an iterable for discovering test cases. Check the
:class:`TestLoader` for details.
In the above example the test suite will look for all python files
in the ``regression`` module (in a recursive fashion), and for modules
called ``tests`` in the ``example`` module.

.. _apps-test-loading:

Loading Tests
=================

Loading test cases is accomplished via the :class:`TestLoader` class. In
this context we refer to an ``object`` as a ``module`` (including a
directory module) or a ``class``.

These are the rules for loading tests:

* Directories that aren't packages are not inspected.
* Any class that is a ``unittest.TestCase`` subclass is collected.
* if an object starts with ``_`` or ``.`` it won't be collected,
  nor will any objects it contains.
* If an object defines a ``__test__`` attribute that does not evaluate to True,
  that object will not be collected, nor will any objects it contains.


Test Case
=============
Only subclasses of  ``unittest.TestCase`` are collected by this application.
When running a test, pulsar looks for two extra method: ``_pre_setup`` and
``_post_teardown``. If the former is available, it is run just before the
``setUp`` method while if the latter is available, it is run
just after the ``tearDown`` method.

.. _unittest2: http://pypi.python.org/pypi/unittest2
.. _mock: http://pypi.python.org/pypi/mock
'''
__test__ = False
import logging
import os
import sys
import time
import inspect

if sys.version_info >= (2,7):
    import unittest
else:   # pragma nocover
    try:
        import unittest2 as unittest
    except ImportError:
        print('To run tests in python 2.6 you need to install\
 the unittest2 package')
        exit(0)

if sys.version_info < (3,3): # pragma nocover
    try:
        import mock
    except ImportError:
        print('To run tests you need to install the mock package')
        exit(0)
else:
    from unittest import mock
    
import pulsar
from pulsar.apps import tasks
from pulsar.utils import events

from .result import *
from .case import *
from .plugins.base import *
from .loader import *
from .utils import *
from .wsgi import *

def dont_run_with_thread(obj):
    c = pulsar.get_actor().cfg.concurrency
    d = unittest.skipUnless(c=='process',
                            'Run only when concurrency is process')
    return d(obj)


class ExitTest(Exception):
    pass


class TestVerbosity(TestOption):
    name = 'verbosity'
    flags = ['--verbosity']
    type = int
    default = 1
    desc = """Test verbosity, 0, 1, 2, 3"""


class TestLabels(TestOption):
    name = "labels"
    nargs = '*'
    validator = pulsar.validate_list
    desc = """Optional test labels to run. If not provided\
 all tests are run.

To see available labels use the -l option."""


class TestSize(TestOption):
    name = 'size'
    flags = ['--size']
    #choices = ('tiny','small','normal','big','huge')
    default = 'normal'
    desc = """Optional test size."""


class TestList(TestOption):
    name = "list_labels"
    flags = ['-l','--list_labels']
    action = 'store_true'
    default = False
    validator = pulsar.validate_bool
    desc = """List all test labels without performing tests."""


test_commands = set()

@pulsar.command(internal=True, ack=False, commands_set=test_commands)
def test_result(client, actor, sender, tag, clsname, result):
    '''Command for sending test results from test workers to the test monitor.'''
    actor.logger.debug('Got a test results from %s.%s', tag, clsname)
    return actor.app.add_result(actor, result)


class TestSuite(tasks.CPUboundServer):
    '''An asynchronous test suite which works like a task queue where each task
is a group of tests specified in a test class.

:parameter modules: An iterable over modules where to look for tests.
    Alternatively it can be a callable returning the iterable over modules.
    For example::

        suite = TestSuite(modules=('regression',
                                   ('examples','tests'),
                                   ('apps','test_*')))
        
        def get_modules(suite):
            ...

        suite = TestSuite(modules=get_modules)

    If not provided it is set as default to ``["tests"]`` which loads all
    python module from the tests module in a recursive fashion.
    Check the the :class:`TestLoader` for detailed information.

:parameter result_class: Optional class for collecting test results. By default
    it used the standard ``unittest.TextTestResult``.
:parameter plugins: Optional list of :class:`TestPlugin` instances.
'''
    _app_name = 'test'
    cfg_apps = ('cpubound',)
    commands_set = test_commands
    cfg = {'loglevel': 'none',
           'timeout': 3600,
           'backlog': 1,
           'logconfig': {
                'loggers': {
                    LOGGER.name: {'handlers': ['console_message'],
                                  'level': logging.INFO}
                            }
                         }
           }

    def handler(self):
        return self

    def python_path(self):
        #Override the python path so that we put the directory where the script
        #is in the ppython path
        path = os.getcwd()
        if path not in sys.path:
            sys.path.insert(0, path)
    
    def on_config_init(self, cfg, params):
        self.plugins = params.get('plugins') or ()
        if self.plugins:
            for plugin in self.plugins:
                cfg.settings.update(plugin.config.settings)
    
    @property
    def runner(self):
        '''Instance of :class:`TestRunner` driving the test case
configuration and plugins.'''
        if 'runner' not in self.local:
            result_class = getattr(self, 'result_class', None)
            r = unittest.TextTestRunner()
            stream = r.stream
            runner = TestRunner(self.plugins, stream, result_class)
            abort_message = runner.configure(self.cfg)
            if abort_message:
                raise ExitTest(str(abort_message))
            self.local.runner = runner
        return self.local.runner

    def on_config(self):
        #When config is available load the tests and check what type of
        #action is required.
        modules = getattr(self, 'modules', None)
        if not hasattr(self, 'plugins'):
            self.plugins = ()
        # Create a runner and configure it
        runner = self.runner
        if not modules:
            modules = ['tests']
        if hasattr(modules, '__call__'):
            modules = modules(self)
        loader = TestLoader(os.getcwd(), modules, runner, logger=self.logger)
        # Listing labels
        if self.cfg.list_labels:
            tags = self.cfg.labels
            if tags:
                s = '' if len(tags) == 1 else 's'
                print('\nTest labels for label{0} {1}\n'\
                      .format(s,', '.join(tags)))
            else:
                print('\nAll test labels\n')
            def _tags():
                for tag, mod in loader.testmodules(tags):
                    doc = mod.__doc__
                    if doc:
                        tag = '{0} - {1}'.format(tag,doc)
                    yield tag
            for tag in sorted(_tags()):
                print(tag)
            print('\n')
            return False
        self.local.loader = loader

    def monitor_start(self, monitor):
        # When the monitor starts load all :class:`TestRequest` into the
        # in the :attr:`pulsar.Actor.ioqueue`.
        loader = self.local.loader
        tags = self.cfg.labels
        try:
            self.local.tests = tests = list(loader.testclasses(tags))
            if tests:
                self.logger.info('loaded %s test classes', len(tests))
                self.runner.on_start()
                events.fire('tests', self, tests=tests)
                monitor.cfg.set('workers', min(self.cfg.workers, len(tests)))
                self._time_start = None
            else:
                raise ExitTest('Could not find any tests.')
        except ExitTest as e:
            print(str(e))
            monitor.arbiter.stop()
        except Exception:
            LOGGER.critical('Error occurred before starting tests',
                            exc_info=True)
            monitor.arbiter.stop()

    def monitor_task(self, monitor):
        if self._time_start is None:
            tests = self.local.tests
            self.logger.info('sending %s test classes to the task queue',
                          len(tests))
            self._time_start = time.time()
            for tag, testcls in tests:
                monitor.put(TestRequest(testcls, tag))

    def add_result(self, monitor, result):
        #Check if we got all results
        runner = self.runner
        runner.add(result)
        if runner.count == len(self.local.tests):
            time_taken = time.time() - self._time_start
            runner.on_end()
            runner.printSummary(time_taken)
            # Shut down the arbiter
            if runner.result.errors or runner.result.failures:
                exit_code = 1
            else:
                exit_code = 0 
            return monitor.arbiter.stop(exit_code=exit_code)




import pulsar
import sys

from .states import REVOKED

__all__ = ['TaskQueueException',
           'TaskException',
           'TaskNotAvailable',
           'TaskTimeout',
           'SchedulingError']

class TaskQueueException(pulsar.PulsarException):
    pass


class TaskException(TaskQueueException):
    task_name = None


class TaskNotAvailable(TaskException):
    MESSAGE = 'Task {0} is not registered. Check your settings.'
    def __init__(self, task_name):
        self.task_name = task_name
        super(TaskNotAvailable,self).__init__(self.MESSAGE.format(task_name))


class TaskTimeout(TaskException):
    status = REVOKED


class SchedulingError(TaskException):
    pass



import time
from datetime import timedelta, datetime

from pulsar.utils.httpurl import itervalues, iteritems
from pulsar.utils.timeutils import remaining, timedelta_seconds,\
                                     humanize_seconds

from .models import registry
from .exceptions import SchedulingError, TaskNotAvailable
from .states import PENDING


__all__ = ['Scheduler']


EMPTY_TUPLE = ()
EMPTY_DICT = {}


def get_datetime(expiry, start):
    if isinstance(expiry, datetime):
        return expiry
    elif isinstance(expiry, timedelta):
        return start + expiry
    else:
        return datetime.fromtimestamp(expiry)


class Schedule(object):

    def __init__(self, run_every=None, anchor=None):
        self.run_every = run_every
        self.anchor = anchor

    def remaining_estimate(self, last_run_at, now=None):
        """Returns when the periodic task should run next as a timedelta."""
        return remaining(last_run_at, self.run_every, now=now)

    def is_due(self, last_run_at, now=None):
        """Returns tuple of two items ``(is_due, next_time_to_run)``,
        where next time to run is in seconds.

        See :meth:`unuk.contrib.tasks.models.PeriodicTask.is_due` for more information.
        """
        rem_delta = self.remaining_estimate(last_run_at, now = now)
        rem = timedelta_seconds(rem_delta)
        if rem == 0:
            return True, timedelta_seconds(self.run_every)
        return False, rem


class SchedulerEntry(object):
    """A class which can be used as a schedule entry in a
:class:`Scheduler` instance."""
    name = None
    '''Task name'''
    schedule = None
    '''The schedule'''
    last_run_at = None
    '''The time and date of when this task was last run.'''
    total_run_count = None
    '''Total number of times this periodic task has been executed.'''

    def __init__(self, name, schedule, args=(), kwargs={},
                 last_run_at = None, total_run_count=None):
        self.name = name
        self.schedule = schedule
        self.last_run_at = last_run_at or datetime.now()
        self.total_run_count = total_run_count or 0

    def __repr__(self):
        return self.name
    __str__ = __repr__

    @property
    def scheduled_last_run_at(self):
        '''The scheduled last run datetime. This is different from :attr:`last_run_at` only when :attr:`anchor` is set.'''
        last_run_at = self.last_run_at
        anchor = self.anchor
        if last_run_at and anchor:
            run_every = self.run_every
            times = int(timedelta_seconds(last_run_at - anchor)\
                            /timedelta_seconds(run_every))
            if times:
                anchor += times*run_every
                while anchor <= last_run_at:
                    anchor += run_every
                while anchor > last_run_at:
                    anchor -= run_every
                self.schedule.anchor = anchor
            return anchor
        else:
            return last_run_at

    @property
    def run_every(self):
        return self.schedule.run_every

    @property
    def anchor(self):
        return self.schedule.anchor

    def next(self, now = None):
        """Returns a new instance of the same class, but with
        its date and count fields updated. Function called by :class:`Scheduler`
        when the ``this`` is due to run."""
        now = now or datetime.now()
        self.last_run_at = now or datetime.now()
        self.total_run_count += 1
        return self

    def is_due(self, now = None):
        return self.schedule.is_due(self.scheduled_last_run_at, now = now)


class Scheduler(object):
    """Scheduler for periodic tasks. This class is the main driver of tasks
and task scheduling.

.. attribute:: task_class

    The :attr:`TaskQueue.task_class` for producing new :class:`Task`.

"""
    def __init__(self, app):
        self._entries = self.setup_schedule()
        self.next_run = datetime.now()
        self.task_class = app.task_class
        self.logger = app.logger

    @property
    def entries(self):
        return self._entries

    def queue_task(self, monitor, jobname, targs=None, tkwargs=None, **params):
        '''Create a new :class:`Task` which may or may not queued.

:parameter monitor: the :class:`pulsar.ApplicationMonitor` running the
    :class:`TaskQueue` application.
:parameter jobname: the name of a :class:`Job` registered
    with the :class:`TaskQueue` application.
:parameter targs: optional tuple used for the positional arguments in the
    task callable.
:parameter tkwargs: optional dictionary used for the key-valued arguments
    in the task callable.
:parameter params: Additional parameters to be passed to the :class:`Task`
    constructor (not its callable function).

:rtype: an instance of :class:`Task`'''
        task = self._make_request(jobname, targs, tkwargs, **params)
        if task.needs_queuing():
            task._queued = True
            monitor.put(task.serialize_for_queue())
        else:
            task._queued = False
            self.logger.debug('Task %s already requested, abort.', task)
        return task

    def tick(self, monitor, now=None):
        '''Run a tick, that is one iteration of the scheduler.
Executes all due tasks calculate the time in seconds to wait before
running a new :meth:`tick`. For testing purposes a :class:`datetime.datetime`
value ``now`` can be passed.'''
        remaining_times = []
        try:
            for entry in itervalues(self._entries):
                is_due, next_time_to_run = entry.is_due(now=now)
                if is_due:
                    self.queue_task(monitor, entry.name)
                if next_time_to_run:
                    remaining_times.append(next_time_to_run)
        except:
            self.logger.error('Error in task scheduler', exc_info=True)
        self.next_run = now or datetime.now()
        if remaining_times:
            self.next_run += timedelta(seconds = min(remaining_times))

    def maybe_schedule(self, s, anchor):
        if isinstance(s, int):
            s = timedelta(seconds=s)
        if not isinstance(s, timedelta):
            raise ValueError('Schedule %s is not a timedelta' % s)
        return Schedule(s, anchor)

    def setup_schedule(self):
        entries = {}
        for name, task in registry.filter_types('periodic'):
            schedule = self.maybe_schedule(task.run_every, task.anchor)
            entries[name] = SchedulerEntry(name, schedule)
        return entries

    def job_list(self, jobnames = None):
        jobnames = jobnames or registry
        for name in jobnames:
            if name not in registry:
                continue
            job = registry[name]
            can_overlap = job.can_overlap
            if hasattr(can_overlap, '__call__'):
                can_overlap = 'maybe'
            d = {'doc':job.__doc__,
                 'doc_syntax':job.doc_syntax,
                 'type':job.type,
                 'can_overlap': can_overlap}
            if name in self.entries:
                entry = self.entries[name]
                _,next_time_to_run = self.next_scheduled((name,))
                run_every = 86400*job.run_every.days + job.run_every.seconds
                d.update({'next_run':next_time_to_run,
                          'run_every':run_every,
                          'runs_count':entry.total_run_count})
            yield (name,d)

    def next_scheduled(self, jobnames=None):
        if jobnames:
            entries = (self._entries.get(name, None) for name in jobnames)
        else:
            entries = itervalues(self._entries)
        next_entry = None
        next_time = None
        for entry in entries:
            if entry is None:
                continue
            is_due, next_time_to_run = entry.is_due()
            if is_due:
                next_time = 0
                next_entry = entry
                break
            elif next_time_to_run is not None:
                if next_time is None or next_time_to_run < next_time:
                    next_time = next_time_to_run
                    next_entry = entry
        if next_entry:
            return (next_entry.name, max(next_time, 0))
        else:
            return (jobnames, None)

    def get_task(self, id, remove=False):
        if isinstance(id, self.task_class):
            task = id
        else:
            task = self.task_class.get_task(self, id)
        if task and task.done() and remove:
            self.delete_tasks([task.id])
        else:
            return task
        
    def get_tasks(self, **parameters):
        return self.task_class.get_tasks(self, **parameters)
    
    def save_task(self, task):
        return self.task_class.save_task(self, task)
    
    def delete_tasks(self, ids=None):
        return self.task_class.delete_tasks(self, ids)
    
    ############################################################################
    ##    PRIVATE METHODS
    ############################################################################

    def _make_request(self, jobname, targs=None, tkwargs=None, expiry=None,
                      **params):
        if jobname in registry:
            task_class = self.task_class
            job = registry[jobname]
            targs = targs or EMPTY_TUPLE
            tkwargs = tkwargs or EMPTY_DICT
            id = job.make_task_id(targs, tkwargs)
            task = self.get_task(id, remove=True)
            if task:
                return task.to_queue(self)
            else:
                if job.name in self.entries:
                    self.entries[job.name].next()
                time_executed = datetime.now()
                if expiry is not None:
                    expiry = get_datetime(expiry, time_executed)
                elif job.timeout:
                    expiry = get_datetime(job.timeout, time_executed)
                task = task_class(id=id, name=job.name,
                                  time_executed=time_executed, expiry=expiry,
                                  args=targs, kwargs=tkwargs, status=PENDING,
                                  **params)
                task.on_created(self)
                return task.to_queue(self)
        else:
            raise TaskNotAvailable(jobname)



import sys
from collections import deque

from pulsar import CommandNotFound, AuthenticationError

from .defer import Deferred, is_async, make_async, AlreadyCalledError
from .mailbox import mailbox, ActorMessage
from .access import get_actor

__all__ = ['ActorMessage',
           'ActorProxyDeferred',
           'ActorProxy',
           'ActorProxyMonitor',
           'get_proxy',
           'command',
           'get_command']

global_commands_table = {}
actor_commands = set()
arbiter_commands = set()

def get_proxy(obj, safe=False):
    if isinstance(obj, ActorProxy):
        return obj
    elif hasattr(obj, 'proxy'):
        return get_proxy(obj.proxy)
    else:
        if safe:
            return None
        else:
            raise ValueError('"%s" is not an actor or actor proxy.' % obj)
            
def get_command(name, commands_set=None):
    '''Get the command function *name*'''
    name = name.lower()
    if commands_set and name not in commands_set:
        return
    return global_commands_table.get(name)


class command:
    '''Decorator for pulsar command functions.
    
:parameter ack: ``True`` if the command acknowledge the sender with a
    response. Usually is set to ``True`` (which is also the default value).
:parameter authenticated: If ``True`` the action can only be invoked by
    remote actors which have authenticated with the actor for which
    the action has been requested.
:parameter internal: Internal commands are for internal use only, not for
    external clients. They accept the actor proxy calling as third positional
    arguments.
'''
    def __init__(self, ack=True, authenticated=False, internal=False,
                 commands_set=None):
        self.ack = ack
        self.authenticated = authenticated
        self.internal = internal
        if commands_set is None:
            commands_set = actor_commands
        self.commands_set = commands_set
    
    def __call__(self, f):
        self.name = f.__name__.lower()
        
        def command_function(client, actor, *args, **kwargs):
            if self.authenticated:
                password = actor.cfg.get('password')
                if password and not client.connection.authenticated:
                    raise AuthenticationError()
            if self.internal:
                caller = get_proxy(args[0]) if args else None
                if not caller:
                    raise AuthenticationError()
                args = args[1:]
                return f(client, actor, caller, *args, **kwargs)
            else:
                return f(client, actor, *args, **kwargs)
        
        command_function.ack = self.ack
        command_function.internal = self.internal
        command_function.__name__ = self.name
        command_function.__doc__ = f.__doc__
        self.commands_set.add(self.name)
        global_commands_table[self.name] = command_function
        return command_function
            
            
class ActorProxyDeferred(Deferred):
    '''A :class:`Deferred` for an :class:`ActorProxy`. This instance will be
obtain and :class:`ActorProxy` result once the remote :class:`Actor` is fully
functional.

.. attribute:: aid

    The the remote :attr:`Actor` id
     
'''
    def __init__(self, aid, msg=None):
        super(ActorProxyDeferred,self).__init__()
        if msg is None:
            # In this case aid is an instance of an ActorProxyMonitor
            proxy = aid
            self.aid = proxy.aid
            # callback for the mailbox
            proxy.on_address = self
        else:
            self.aid = aid
            # simply listent for the calbacks and errorbacks
            msg.addBoth(self.callback)
    
    def __repr__(self):
        return '%s(%s)' % (self.__class__, self.aid)
    __str__ = __repr__
    
    
class ActorProxy(object):
    '''This is an important component in pulsar concurrent framework. An
instance of this class is as a proxy for a remote `underlying` 
:class:`Actor`. This is a lightweight class which delegates
function calls to the underlying remote object.

It is pickable and therefore can be send from actor to actor using pulsar
messaging. It exposes all the underlying :class:`command` which have been
implemented.

For example, lets say we have a proxy ``a``, to send a message to it::

    from pulsar import send
    
    send(a, 'echo', 'hello there!')
    
will send the command ``echo`` to actor ``a`` with
parameter ``"hello there!"``.
    
.. attribute:: aid

    Unique ID for the remote :class:`Actor`
    
.. attribute:: address

    the socket address of the underlying :attr:`Actor.mailbox`.
    
'''
    def __init__(self, impl):
        self.aid = impl.aid
        self.name = impl.name
        self.commands_set = impl.commands_set
        self.address = impl.address
        self.cfg = impl.cfg
    
    def __repr__(self):
        return '%s(%s)' % (self.name, self.aid)
    __str__ = __repr__
    
    @property
    def mailbox(self):
        '''Actor mailbox'''
        if self.address:
            return get_actor().proxy_mailbox(self.address)
        
    @property
    def proxy(self):
        return self
    
    def receive_from(self, sender, command, *args, **kwargs):
        '''Send an :class:`ActorMessage` to the underlying actor
(the receiver). This is the low level function call for
communicating between actors.

:parameter sender: :class:`Actor` sending the message.
:parameter command: the :class:`command` to perform in the actor underlying
    this proxy.
:parameter args: non positional arguments of command.
:parameter kwargs: key-valued arguments of command.
:rtype: an asynchronous :class:`ActorMessage`.'''
        if sender is None:
            sender = get_actor()
        if not self.mailbox:
            sender.logger.critical('Cannot send a message to %s. No\
 mailbox available.', self)
            return
        cmd = get_command(command, self.commands_set)
        if not cmd:
            raise CommandNotFound(command)
        msg = ActorMessage(cmd.__name__, sender, self.aid, args, kwargs)
        send = self.mailbox.execute if cmd.ack else self.mailbox.send
        return send(msg)
    
    def __eq__(self, o):
        o = get_proxy(o,True)
        return o and self.aid == o.aid
    
    def __ne__(self, o):
        return not self.__eq__(o)
        
    def stop(self, sender=None):
        '''Stop the remote :class:`Actor`'''
        self.receive_from(sender, 'stop')
        

class ActorProxyMonitor(ActorProxy):
    '''A specialised :class:`ActorProxy` class which contains additional
information about the remote underlying :class:`pulsar.Actor`. Unlike the
:class:`pulsar.ActorProxy` class, instances of this class are not pickable and
therefore remain in the :class:`Arbiter` process domain, which is the
process where they have been created.

.. attribute:: impl

    The :class:`Concurrency` instance for the remote :class:`Actor`.
    
.. attribute:: info

    Dictionary of information regarding the remote :class:`Actor`
'''
    monitor = None
    def __init__(self, impl):
        self.impl = impl
        self.info = {}
        self.stopping_loops = 0
        super(ActorProxyMonitor,self).__init__(impl)
        
    @property
    def notified(self):
        return self.info.get('last_notified')
    
    @property
    def pid(self):
        return self.impl.pid
    
    @property
    def proxy(self):
        return ActorProxy(self)
    
    def is_alive(self):
        '''True if underlying actor is alive'''
        return self.impl.is_alive()
        
    def terminate(self):
        '''Terminate life of underlying actor.'''
        self.impl.terminate()
    
    def join(self, timeout=None):
        '''Wait until the underlying actor terminates. If *timeout* is
provided, it raises an exception if the timeout is reached.'''
        self.impl.join(timeout=timeout)

    def start(self):
        '''Start the remote actor.'''
        self.impl.start()

    


from datetime import datetime, date, timedelta
from hashlib import sha1
import logging
import inspect

from pulsar.utils.httpurl import iteritems
from pulsar.utils.security import gen_unique_id


__all__ = ['JobMetaClass','Job','PeriodicJob',
           'anchorDate','JobRegistry','registry',
           'create_task_id']


def create_task_id():
    return gen_unique_id()[:8]


class JobRegistry(dict):
    """Site registry for tasks."""

    def regular(self):
        """A generator of all regular jobs."""
        return self.filter_types("regular")

    def periodic(self):
        """A generator of all periodic jobs."""
        return self.filter_types("periodic")

    def register(self, job):
        """Register a job in the job registry.

        The task will be automatically instantiated if not already an
        instance.

        """
        job = inspect.isclass(job) and job() or job
        name = job.name
        self[name] = job

    def filter_types(self, type):
        """Return a generator of all tasks of a specific type."""
        return ((job_name, job)
                    for job_name, job in iteritems(self)
                            if job.type == type)


registry = JobRegistry()



class JobMetaClass(type):
    """Metaclass for Jobs. It performs a little ammount of magic
by:

* Automatic registration of :class:`Job` instances to the
  global :class:`JobRegistry`, unless
  the :attr:`Job.abstract` attribute is set to ``True``.
* If no :attr:`Job.name`` attribute is provided,
  it is automatically set to the class name in lower case.
* Add a logger instance with name given by 'job.name` where name is the
  same as above.
"""

    def __new__(cls, name, bases, attrs):
        super_new = super(JobMetaClass, cls).__new__
        job_module = attrs.get("__module__",None)

        # Abstract class, remove the abstract attribute so
        # any class inheriting from this won't be abstract by default.
        if attrs.pop("abstract", None) or not attrs.get("autoregister", True)\
                                       or not job_module:
            return super_new(cls, name, bases, attrs)
        # Automatically generate missing name.
        job_name = attrs.get("name",name).lower()
        log_prefix = attrs.get("log_prefix") or "pulsar"
        # Because of the way import happens (recursively)
        # we may or may not be the first time the Job tries to register
        # with the framework. There should only be one class for each Job
        # name, so we always return the registered version.
        if job_name not in registry:
            attrs["name"] = job_name
            logname = '%s.job.%s' % (log_prefix, name)
            attrs['logger'] = logging.getLogger(logname)
            job_cls = super_new(cls, name, bases, attrs)
            registry.register(job_cls)
        return registry[job_name].__class__


JobBase = JobMetaClass('JobBase',(object,),{'abstract':True})


class Job(JobBase):
    '''The Job class which is used in a distributed task queue.

.. attribute:: name

    The unique name which defines the Job and which can be used to retrieve
    it from the job registry.

.. attribute:: abstract

    If set to ``True`` (default is ``False``), the Job won't be registered
    with the :class:`JobRegistry`.

.. attribute:: autoregister

    If ``False`` (default is ``True``), the Job need to be registered
    manually with the Job registry.

.. attribute:: type

    Type of Job, one of ``regular`` and ``periodic``.

.. attribute:: timeout

    An instance of a datetime.timedelta or ``None``. If set, it represents the
    time lag after which a task which did not start expires.

    Default: ``None``.

.. attribute:: can_overlap

    Boolean indicating if this job can generate overlapping tasks. It can
    also be a callable which accept the same input parameters as the job
    callable function.

    Default: ``True``.

.. attribute:: doc_syntax

    The doc string syntax.

    Default: ``markdown``

.. attribute:: loglevel

    Level of task logging

    Default: ``None``

.. attribute:: logger

    an instance of a logger.
'''
    abstract = True
    autoregister = True
    type = "regular"
    timeout = None
    expires = None
    loglevel = None
    logformatter = None
    doc_syntax = 'markdown'
    can_overlap = True
    _ack = True

    def __call__(self, consumer, *args, **kwargs):
        '''The Jobs' task executed by the consumer. This function needs to be
implemented by subclasses.'''
        raise NotImplementedError("Jobs must define the run method.")

    def make_task_id(self, args, kwargs):
        '''Get the task unique identifier.
This can be overridden by Job implementation.

:parameter args: tuple of positional arguments passed to the job callable.
:parameter kwargs: dictionary of key-valued parameters passed to
    the job callable.
:rtype: a native string.

Called by the :attr:`TaskQueue.scheduler` when creating a new task.
'''
        can_overlap = self.can_overlap
        if hasattr(can_overlap,'__call__'):
            can_overlap = can_overlap(*args, **kwargs)
        if can_overlap:
            return create_task_id()
        else:
            suffix = ''
            if args:
                suffix = ' args(%s)' % ', '.join((str(a) for a in args))
            if kwargs:
                suffix += ' kwargs(%s)' % ', '.join(('%s=%s' % (k, kwargs[k])\
                            for k in sorted(kwargs)))
            name = '%s%s' % (self.name, suffix)
            return sha1(name.encode('utf-8')).hexdigest()[:8]

    def on_same_id(self, task):
        '''Callback invocked when a task has an id equal to a task already
submitted to the task queue. By default return None and the
task is aborted.'''
        return None

    def ack(self, args, kwargs):
        '''Return ``True`` if a Job task will acknowledge the task queue
 once the result is ready or an error has occured.
 By default it returns ``True`` but it can be overridden so that its
 behaviour can change at runtime.'''
        return self._ack

    def send_to_queue(self, consumer, jobname, *args, **kwargs):
        '''Send a new task request to the :class:`TaskQueue`
from within another :class:`Task`.
This allows tasks to act as tasks factories.

:parameter consumer: The :class:`TaskConsumer` handling the :class:`Task`.
:parameter jobname: The name of the :class:`Job` to run.
:parameter args: positional argument for the :ref:`job callable <job-callable>`.
:parameter kwargs: key-valued parameters for the
    :ref:`job callable <job-callable>`.
:rtype: A :class:`pulsar.ActorMessage` if **ack=True** was passed in the
    key-valued parameters, otherwise nothing.
'''
        worker = consumer.worker
        ack = kwargs.pop('ack', False)
        oper = "addtask" if ack else "addtask_noack"
        res = worker.send(worker.monitor, oper, jobname,
                          {'from_task': consumer.task.id},
                           *args, **kwargs)
        return res if ack else None


class PeriodicJob(Job):
    '''A periodic :class:`Job` implementation.'''
    abstract  = True
    type      = "periodic"
    anchor    = None
    '''If specified it must be a :class:`datetime.datetime` instance.
It controls when the periodic Job is run.'''
    run_every = None
    '''Periodicity as a :class:`datetime.timedelta` instance.'''

    def __init__(self, run_every = None):
        self.run_every = run_every or self.run_every
        if self.run_every is None:
            raise NotImplementedError('Periodic Jobs must have\
 a run_every attribute set, "{0}" does not have one'.format(self.name))

    def is_due(self, last_run_at):
        """Returns tuple of two items ``(is_due, next_time_to_run)``,
where next time to run is in seconds. e.g.

* ``(True, 20)``, means the Job should be run now, and the next
    time to run is in 20 seconds.

* ``(False, 12)``, means the Job should be run in 12 seconds.

You can override this to decide the interval at runtime.
        """
        return self.run_every.is_due(last_run_at)



def anchorDate(hour = 0, minute = 0, second = 0):
    td = date.today()
    return datetime(year = td.year, month = td.month, day = td.day,
                    hour = hour, minute = minute, second = second)





import logging
import os
import sys
from inspect import isfunction

import pulsar
from pulsar import Actor, make_async, safe_async, is_failure, HaltServer,\
                     Monitor, loop_timeout, Deferred, get_actor, async
from pulsar.async.defer import pickle
from pulsar.utils import events
from pulsar.utils.log import LogInformation

__all__ = ['Application',
           'MultiApp',
           'Worker',
           'ApplicationMonitor',
           'get_application']

APP_EVENTS = ('start', 'stop')

def app_events(app):
    for name in APP_EVENTS:
        yield name, Deferred(description='%s.%s' % (app.name, name))
        
def fire_event(f):
    name = f.__name__[3:]
    if name not in APP_EVENTS:
        raise ValueError()
    
    def fire(result, app):
        if not is_failure(result):
            result = app
        else:
            result.log()
        events.fire(name, app)
        app.events[name].callback(app)
        
    def _(*args, **kwargs):
        app = args[0].params.app
        safe_async(f, args, kwargs).addBoth(lambda r: fire(r, app))
    
    _.__name__ = f.__name__
    _.__doc__ = f.__doc__
    return _
            
def get_application(name):
    '''Invoked in the arbiter domain, this function will return
the :class:`Application` associated with *name* if available. If not in the
:class:`Arbiter` domain it returns nothing.'''
    actor = get_actor()
    if actor and actor.is_arbiter():
        monitor = actor.monitors.get(name)
        if monitor:
            return getattr(monitor, 'app', None) 


class Worker(Actor):
    """\
An :class:`Actor` for serving a pulsar :class:`Application`.

.. attribute:: app

    Instance of the :class:`Application` to be performed by the worker

.. attribute:: cfg

    Configuration dictionary

.. attribute:: app_handler

    The application handler obtained from :meth:`Application.handler`.

"""
    @property
    def app(self):
        return self.params.app

    # Delegates Callbacks to the application
    def on_start(self):
        app = self.app
        self.app_handler = app.handler()
        self.app.worker_start(self)
        try:
            self.cfg.worker_start(self)
        except:
            pass

    def on_stop(self):
        return self.app.worker_stop(self)

    def on_exit(self):
        self.app.worker_exit(self)
        try:
            self.cfg.worker_exit(self)
        except:
            pass

    def on_info(self, info):
        return self.app.on_info(self, info)
    
    def on_event(self, fd, event):
        return self.app.on_event(self, fd, event)


class ApplicationMonitor(Monitor):
    '''A :class:`Monitor` for managing a pulsar :class:`Application`.
    
.. attribute:: app_handler

    The monitor application handler obtained from
    :meth:`Application.monitor_handler`.
'''
    actor_class = Worker

    @property
    def app(self):
        return self.params.app
        
    ############################################################################
    # Delegates Callbacks to the application
    @fire_event
    def on_start(self):
        super(ApplicationMonitor, self).on_start()
        app = self.app
        app.monitor_start(self)
        # If no workers available invoke the worker start method too
        if not self.cfg.workers:
            self.app_handler = app.handler()
            app.worker_start(self)
        else:
            self.app_handler = app.monitor_handler()

    @fire_event
    def on_stop(self):
        app = self.params.app
        if not self.cfg.workers:
            yield app.worker_stop(self)
        yield app.monitor_stop(self)
        yield super(ApplicationMonitor, self).on_stop()

    def on_exit(self):
        app = self.params.app
        app.monitor_exit(self)
        try:
            self.cfg.worker_exit(self)
        except:
            pass
    
    def monitor_task(self):
        self.app.monitor_task(self)
        
    def on_event(self, fd, event):
        return self.app.on_event(self, fd, event)
    
    def actorparams(self):
        p = Monitor.actorparams(self)
        app = self.params.app
        if self.cfg.concurrency == 'thread':
            app = pickle.loads(pickle.dumps(app))
        p.update({'app': app,
                  'name': '{0}-worker'.format(app.name)})
        return app.actorparams(self, p)

    def on_info(self, info):
        if not self.cfg.workers:
            return self.params.app.on_info(self, info)
        else:
            return info


class Application(pulsar.Pulsar):
    """An application interface for configuring and loading
the various necessities for any given server or distributed application running
on :mod:`pulsar` concurrent framework.
Applications can be of any sorts or forms and the library is shipped with
several battery included examples in the :mod:`pulsar.apps` framework module.

These are the most important facts about a pulsar :class:`Application`

* Instances must be pickable. If non-pickable data needs to be add on an
  :class:`Application` instance, it must be stored on the
  :attr:`Application.local` dictionary.
* When a new :class:`Application` is initialized,
  a new :class:`ApplicationMonitor` instance is added to the
  :class:`Arbiter`, ready to perform its duties.

:parameter callable: Initialise the :attr:`Application.callable` attribute.
:parameter description: A string describing the application.
    It will be displayed on the command line.
:parameter epilog: Epilog string you will see when interacting with the command
    line.
:parameter name: Application name. If not provided the class name in lower
    case is used
:parameter commands_set: Initialise the :attr:`commands_set` attribute.
:parameter params: a dictionary of configuration parameters which overrides
    the defaults and the `cfg` attribute. They will be overritten by
    a config file or command line arguments.

.. attribute:: app

    A string indicating the application namespace for configuration parameters.

    Default: ``None``.

.. attribute:: callable

    A callable serving your application. The callable must be pickable,
    therefore it is either a function
    or a pickable object. If not provided, the application must
    implement the :meth:`handler` method.

    Default ``None``

.. attribute:: cfg

    dictionary of default configuration parameters.

    Default: ``{}``.

.. attribute:: cfg_apps

    Optional tuple containing names of configuration namespaces to
    be included in the application config dictionary.

    Default: ``None``
    
.. attribute:: script

    full path of the script which starts the application or ``None``.
    If supplied it is used to setup the python path

.. attribute:: commands_set

    Optional set of :ref:`remote actions <api-remote_commands>` available
    on :class:`Actor` created by this :class:`Application`.

    Default: ``None``.
"""
    cfg = {}
    _app_name = None
    description = None
    epilog = None
    cfg_apps = None
    config_options_include = None
    config_options_exclude = None
    commands_set = None

    def __init__(self,
                 callable=None,
                 description=None,
                 name=None,
                 epilog=None,
                 argv=None,
                 script=None,
                 version=None,
                 parse_console=True,
                 commands_set=None,
                 cfg=None,
                 **kwargs):
        '''Initialize a new :class:`Application` and add its
:class:`ApplicationMonitor` to the class:`pulsar.Arbiter`.

:parameter version: Optional version number of the application.

    Default: ``pulsar.__version__``

:parameter parse_console: flag for parsing console inputs. By default it parse
    only if the arbiter has not yet started.
'''
        self.description = description or self.description
        self.epilog = epilog or self.epilog
        self._app_name = self._app_name or self.__class__.__name__.lower()
        self._name = name or self._app_name
        if commands_set is not None:
            self.commands_set = commands_set
        self.script = script
        self.python_path()
        params = cfg or {}
        if self.cfg:
            params.update(self.cfg)
        params.update(kwargs)
        self.callable = callable
        self.load_config(argv, version, parse_console, params)
        self()

    def __call__(self, actor=None):
        if actor is None:
            actor = get_actor()
        monitor = None
        if actor and actor.is_arbiter():
            monitor = actor.monitors.get(self.name)
        if monitor is None and (not actor or actor.is_arbiter()): 
            # Add events
            self.local.events = dict(app_events(self))
            self.cfg.on_start()
            self.configure_logging()
            events.fire('ready', self)
            arbiter = pulsar.arbiter(cfg=self.cfg.new_config())
            if self.on_config() is not False:
                monitor = arbiter.add_monitor(ApplicationMonitor,
                                              self.name,
                                              app=self,
                                              cfg=self.cfg,
                                              ioqueue=self.ioqueue)
                self.cfg = monitor.cfg
                if self.commands_set:
                    monitor.impl.commands_set.update(self.commands_set)
        if self.events:
            return self.events['start']

    @property
    def app(self):
        return self
    
    @property
    def app_name(self):
        return self._app_name

    @property
    def name(self):
        '''Application name, It is unique and defines the application.'''
        return self._name

    @property
    def events(self):
        return self.local.events
    
    def __repr__(self):
        return self.name

    def __str__(self):
        return self.__repr__()

    @property
    def monitor(self):
        actor = get_actor()
        if actor:
            return actor.monitors.get(self.name)
         
    @property
    def ioqueue(self):
        if 'queue' not in self.local:
            self.local.queue = self.get_ioqueue()
        return self.local.queue

    def handler(self):
        '''Returns the callable application handler which is stored in
:attr:`Worker.app_handler`, used by :class:`Worker` to carry out its task.
By default it returns the :attr:`Application.callable`.'''
        return self.callable

    def request_instance(self, worker, fd, event):
        '''Build a request class from a file descriptor *fd* and an *event*.
The returned request instance is passed to the :meth:`handle_request`
method.'''
        return event

    def handle_request(self, worker, request):
        '''This is the main function which needs to be implemented
by actual applications. It is called by the *worker* to handle
a *request*.

:parameter worker: the :class:`Worker` handling the request.
:parameter request: an application specific request object.
:rtype: It can be a generator, a :class:`Deferred` instance
    or the actual response.'''
        raise NotImplementedError()

    def get_ioqueue(self):
        '''Returns an I/O distributed queue for the application if one
is needed. If a queue is returned, the application :class:`Worker`
will have a :class:`IOLoop` instance based on the queue (via :class:`IOQueue`).

By default it returns ``None``.'''
        return None

    def on_config_init(self, cfg, params):
        '''Callback when configuration is initialised but not yet loaded.
This is a chance to add extra :ref:`config parameters <settings>` or remove
unwanted ones. It returns a new :class:`Config` instance or ``None``.'''
        pass
    
    def on_config(self):
        '''Callback when configuration is loaded. This is a chance to do
 an application specific check before the concurrent machinery is put into
 place. If it returns ``False`` the application will abort.'''
        pass

    def python_path(self):
        #Insert the application directory at the top of the python path.
        fname = self.script or os.getcwd()
        path = os.path.split(fname)[0]
        if path not in sys.path:
            sys.path.insert(0, path)

    def add_timeout(self, deadline, callback):
        self.arbiter.ioloop.add_timeout(deadline, callback)
    
    def load_config(self, argv, version, parse_console, params):
        '''Load the application configuration from a file and/or
from the command line. Called during application initialization.

:parameter argv: list of command line parameters to parse.
:parameter version: The version of this application.
:parameter parse_console: True if the console parameters need parsing.
:parameter params: dictionary of parameters passed during construction.

The parameters overriding order is the following:

 * default parameters.
 * the *params* passed in the initialization.
 * the parameters in the optional configuration file
 * the parameters passed in the command line.
'''
        cfg_apps = set(self.cfg_apps or ())
        cfg_apps.add(self.app_name)
        self.cfg_apps = cfg_apps
        cfg = pulsar.Config(self.description,
                            self.epilog,
                            version,
                            self.cfg_apps,
                            self.config_options_include,
                            self.config_options_exclude)
        self.cfg = self.on_config_init(cfg, params)
        if not isinstance(self.cfg, pulsar.Config):
            self.cfg = cfg
        overrides = {}
        specials = set()
        # get the actor if available and override default cfg values with those
        # from the actor
        actor = get_actor()
        if actor and actor.running:
            # actor available and running. unless argv is set, skip parsing
            if argv is None:
                parse_console = False
            for k, v in actor.cfg.items():
                if v is not None:
                    k = k.lower()
                    try:
                        self.cfg.set(k, v)
                        self.cfg.settings[k].default = v
                    except AttributeError:
                        pass
        # modify defaults and values of cfg with params
        for k, v in params.items():
            if v is not None:
                k = k.lower()
                try:
                    self.cfg.set(k, v)
                    self.cfg.settings[k].default = v
                except AttributeError:
                    if not self.add_to_overrides(k, v, overrides):
                        setattr(self, k, v)
        # parse console args
        if parse_console:
            parser = self.cfg.parser()
            opts = parser.parse_args(argv)
            config = getattr(opts, 'config', None)
            if config is not None and self.cfg.config:
                self.cfg.config = config
        else:
            parser, opts = None, None
        #
        # Load up the config file if its found.
        for k, v in self.cfg.import_from_module():
            self.add_to_overrides(k, v, overrides)
        #
        # Update the configuration with any command line settings.
        if opts:
            for k, v in opts.__dict__.items():
                if v is None:
                    continue
                self.cfg.set(k.lower(), v)
        # Lastly, update the configuration with overrides
        for k,v in overrides.items():
            self.cfg.set(k, v)

    def add_to_overrides(self, name, value, overrides):
        names = name.split('__')
        if len(names) == 2 and names[0] == self.name:
            name = names[1].lower()
            if name in self.cfg.settings:
                overrides[name] = value
                return True

    def monitor_handler(self):
        '''Returns an application handler for the :class:`ApplicationMonitor`.
By default it returns ``None``.'''
        return None

    # MONITOR AND WORKER CALLBACKS
    def on_info(self, worker, data):
        return data

    def on_event(self, worker, fd, events):
        pass
    
    # WORKERS CALLBACKS
    def worker_start(self, worker):
        '''Called by the :class:`Worker` :meth:`pulsar.Actor.on_start`
:ref:`callback <actor-callbacks>` method.'''
        pass

    def worker_stop(self, worker):
        '''Called by the :class:`Worker` :meth:`pulsar.Actor.on_stop`
:ref:`callback <actor-callbacks>` method.'''
        pass

    def worker_exit(self, worker):
        '''Called by the :class:`Worker` :meth:`pulsar.Actor.on_exit`
:ref:`callback <actor-callbacks>` method.'''
        pass

    # MONITOR CALLBACKS
    def actorparams(self, monitor, params):
        '''A chance to override the dictionary of parameters *params*
before a new :class:`Worker` is spawned.'''
        return params

    def monitor_start(self, monitor):
        '''Callback by :class:`ApplicationMonitor` when starting.
The application is now in the arbiter but has not yet started.'''
        pass

    def monitor_stop(self, monitor):
        '''Callback by :class:`ApplicationMonitor` when stopping'''
        pass

    def monitor_exit(self, monitor):
        '''Callback by :class:`ApplicationMonitor` at exit'''
        pass

    def monitor_task(self, monitor):
        '''Callback by :class:`ApplicationMonitor` at each event loop'''
        pass
    
    def start(self):
        '''Start the application if it wasn't already started.'''
        arbiter = pulsar.arbiter()
        if arbiter and self.name in arbiter.monitors:
            arbiter.start()
        return self


class MultiApp:
    
    def __init__(self, name='taskqueue', **params):
        self.name = name
        self.params = params
        self.apps = []
        
    def __call__(self, actor=None):
        raise NotImplementedError()
        
    def start(self):
        for app in self.apps:
            app.start()


from pulsar import send, make_async

from .jsonrpc import JSONRPC


__all__ = ['PulsarServerCommands']


class PulsarServerCommands(JSONRPC):
    '''Some useful commands to add to your :class:`JSONRPC` handler to get you
 started. It exposes the following functions:
 
.. method:: ping()

    Return ``pong``.
    
.. method:: server_info()

    Return a dictionary of information regarding the server and workers.
    It invokes the :meth:`extra_server_info` for adding custom information.
    
    :rtype: ``dict``.
    
.. method:: functions_list()

    Return the list of functions available in the rpc handler
    
    :rtype: ``list``.
    
.. method:: kill_actor(aid)

    :parameter aid: :attr:`pulsar.Actor.aid` of the actor to kill
'''
    
    def rpc_ping(self, request):
        '''Ping the server'''
        return 'pong'
    
    def rpc_echo(self, request, message=''):
        '''Echo the server'''
        return message
    
    def rpc_server_info(self, request):
        '''Dictionary of information about the server'''
        info = send('arbiter', 'info')
        return info.add_callback(lambda res: self.extra_server_info(
                                                       request, res))
    
    def rpc_functions_list(self, request):
        return list(self.listFunctions())
    
    def rpc_documentation(self, request):
        '''Documentation in restructured text.'''
        return self.docs()
    
    def rpc_kill_actor(self, request, aid):
        '''Kill and actor which match the id *aid*'''
        return send('arbiter', 'kill_actor', aid)
        
    def extra_server_info(self, request, info):
        '''Add additional information to the info dictionary.'''
        return info
    



'''Original code from https://github.com/benoitc/http-parser

2011 (c) Benoît Chesneau <benoitc@e-engura.org>

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

Modified and adapted for pulsar.
'''
import os
import sys
from http cimport *

ispy3k = sys.version_info[0] == 3

if ispy3k:
    from urllib.parse import urlsplit
else:
    from urlparse import urlsplit

    
class _ParserData:

    def __init__(self, decompress=False):
        self.url = ""
        self.body = []
        self.headers = {}
        self.environ = {}
        
        self.decompress = decompress
        self.decompressobj = None

        self.chunked = False

        self.headers_complete = False
        self.partial_body = False
        self.message_begin = False
        self.message_complete = False
        
        self._last_field = ""
        self._last_was_value = False
        
        
cdef class HttpParser:
    """Cython HTTP parser wrapping http_parser."""

    cdef http_parser _parser
    cdef http_parser_settings _settings
    cdef object _data

    cdef str _path
    cdef str _query_string
    cdef str _fragment
    cdef object _parsed_url

    def __init__(self, kind=2, decompress=False):
        """ constructor of HttpParser object.
:attr kind: Int,  could be 0 to parseonly requests, 
1 to parse only responses or 2 if we want to let
the parser detect the type. 
        """
        # set parser type
        if kind == 2:
            parser_type = HTTP_BOTH
        elif kind == 1:
            parser_type = HTTP_RESPONSE
        elif kind == 0:
            parser_type = HTTP_REQUEST

        # initialize parser
        http_parser_init(&self._parser, parser_type)
        self._data = _ParserData(decompress=decompress)
        self._parser.data = <void *>self._data
        self._parsed_url = None
        self._path = ""
        self._query_string = ""
        self._fragment = ""

        # set callback
        self._settings.on_url = <http_data_cb>on_url_cb
        self._settings.on_body = <http_data_cb>on_body_cb
        self._settings.on_header_field = <http_data_cb>on_header_field_cb
        self._settings.on_header_value = <http_data_cb>on_header_value_cb
        self._settings.on_headers_complete = <http_cb>on_headers_complete_cb
        self._settings.on_message_begin = <http_cb>on_message_begin_cb
        self._settings.on_message_complete = <http_cb>on_message_complete_cb

    def execute(self, char *data, size_t length):
        """ Execute the parser with the last chunk. We pass the length
        to let the parser know when EOF has been received. In this case
        length == 0.

        :return recved: Int, received length of the data parsed. if
        recvd != length you should return an error.
        """
        return http_parser_execute(&self._parser, &self._settings,
                data, length)

    def get_version(self):
        """ get HTTP version """
        return (self._parser.http_major, self._parser.http_minor)

    def get_method(self):
        """ get HTTP method as string"""
        return http_method_str(<http_method>self._parser.method)

    def get_status_code(self):
        """ get status code of a response as integer """
        return self._parser.status_code

    def get_url(self):
        """ get full url of the request """
        return self._data.url

    def maybe_parse_url(self):
        raw_url = self.get_url()
        if not self._parsed_url and raw_url:
            parts = urlsplit(raw_url)
            self._parsed_url = parts
            self._path = parts.path or ""
            self._query_string = parts.query or ""
            self._fragment = parts.fragment or ""

    def get_path(self):
        """ get path of the request (url without query string and
        fragment """
        self.maybe_parse_url()
        return self._path

    def get_query_string(self):
        """ get query string of the url """
        self.maybe_parse_url()
        return self._query_string

    def get_fragment(self):
        """ get fragment of the url """
        self.maybe_parse_url()
        return self._fragment

    def get_headers(self):
        """get request/response headers dictionary."""
        return self._data.headers
    
    def get_protocol(self):
        return None
    
    def get_body(self):
        return self._data.body

    def recv_body(self):
        """ return last chunk of the parsed body"""
        body = b''.join(self._data.body)
        self._data.body = []
        self._data.partial_body = False
        return body

    def recv_body_into(self, barray):
        """ Receive the last chunk of the parsed body and store the data
        in a buffer rather than creating a new string. """
        l = len(barray)
        body = b''.join(self._data.body)
        m = min(len(body), l)
        data, rest = body[:m], body[m:]
        barray[0:m] = data
        if not rest:
            self._data.body = []
            self._data.partial_body = False
        else:
            self._data.body = [rest]
        return m

    def is_headers_complete(self):
        """ return True if all headers have been parsed. """ 
        return self._data.headers_complete

    def is_partial_body(self):
        """ return True if a chunk of body have been parsed """
        return self._data.partial_body

    def is_message_begin(self):
        """ return True if the parsing start """
        return self._data.message_begin

    def is_message_complete(self):
        """ return True if the parsing is done (we get EOF) """
        return self._data.message_complete

    def is_chunked(self):
        """ return True if Transfer-Encoding header value is chunked"""
        te = self._data.headers.get('transfer-encoding', '').lower()
        return te == 'chunked'



'''Asynchronous application for serving requests
on a socket. This is the base class of :class:`pulsar.apps.wsgi.WSGIServer`.

A simple echo server ``script.py`` can be implemented as follow::

    import pulsar
    from pulsar.apps.socket import SocketServer
    
    class EchoProtocol:
        "The simplest protocol possible."
        def decode(self, data):
            return bytes(data), bytearray()
        
        def encode(self, data):
            return data
    
    def echoserver(actor, socket, **kwargs):
        kwargs['protocol_factory'] = EchoProtocol
        return pulsar.AsyncSocketServer(actor, socket, **kwargs)
        
    if __name__ == '__main__':
        SocketServer(socket_server_factory=echoserver).start()
    
    
The main parameter for :class:`SocketServer` is the ``socket_server_factory``
callable which build a :class:`pulsar.AsyncSocketServer` for each
:class:`pulsar.Worker` serving the application.

.. _socket-protocol:

Protocol
==============

The protocol class needs to implement two methods, the ``decode`` for servers
and the ``encode`` for clients. Both methods accept as input one parameter which
can be ``bytes`` or a ``bytesarray``.

The ``decode`` method must return a **two elements tuple** of the form:

 * ``None``, ``bytesarray``: when more data is needed to form a message.
 * ``message``, ``bytesarray``: where *message* is a correctly parsed message
   and *bytesarray* contains data for the next message. 

The ``encode`` method must return ``bytes`` which will be sent to the server.

Useful settings
==================
All standard :ref:`settings <settings>` can be applied to a
:class:`SocketServer`. These are some of the most important:

bind
------
The :class:`SocketServer` application introduces the additional pulsar settings
:ref:`bind <setting-bind>` which is used to specify the address
to bind the server to::

    python script.py --bind 127.0.0.1:8070

backlog
---------   
To control the concurrency of the server you can use the
:ref:`backlog <setting-backlog>` settings. For example::

    python script.py --backlog 1000
    
will serve a maximum of 1000 clients per :class:`pulsar.Worker` concurrently.

keepalive
---------------
To control how long a :class:`pulsar.AsyncConnection` is kept alive after the
last read from the remote client, one can use the
:ref:`keepalive <setting-keepalive>` setting.

    python script.py --keepalive 10
    
will close a client connection which has been idle for 10 seconds.
 


Concurrency
==================

When running a :class:`SocketServer` in threading mode::

    python script.py --concurrency thread
 
the number of :class:`pulsar.Worker` serving the application is set to 0 so
that the application is actually served by the :class:`Arbiter`
eventloop (we refer this to a single process server).
This configuration is used when debugging, testing, benchmarking or small
load servers.

In addition, a :class:`SocketServer` in multi-process mode is only available for:
    
 * Posix systems
 * Windows running python 3.2 or above.
'''
import pulsar
from pulsar import AsyncIOStream


class Bind(pulsar.Setting):
    section = "Socket Servers"
    app = 'socket'
    name = "bind"
    flags = ["-b", "--bind"]
    meta = "ADDRESS"
    default = "127.0.0.1:{0}".format(pulsar.DEFAULT_PORT)
    desc = """\
        The socket to bind.
        
        A string of the form: 'HOST', 'HOST:PORT', 'unix:PATH'. An IP is a valid
        HOST.
        """
        
class Keepalive(pulsar.Setting):
    section = "Socket Servers"
    app = 'socket'
    name = "keepalive"
    flags = ["--keep-alive"]
    validator = pulsar.validate_pos_int
    type = int
    default = 15
    desc = """\
        The number of seconds to keep an idle client connection
        open."""


class SocketServer(pulsar.Application):
    '''This application bind a socket to a given address and listen for
requests. The request handler is constructued from the
:attr:`socket_server_factory` parameter/attribute.
    
.. attribute:: socket_server_factory

    Callable which returns the asynchronous socket server, usually
    a subclass of :class:`pulsar.AsyncSocketServer`.
    
.. attribute:: address

    The socket address, available once the application has started.
    '''
    _app_name = 'socket'
    socket_server_factory = None
    address = None
    
    def monitor_start(self, monitor):
        # if the platform does not support multiprocessing sockets set
        # the number of workers to 0.
        cfg = self.cfg
        if not self.socket_server_factory:
            raise TypeError('Socket server class not specified.')
        if not pulsar.platform.has_multiProcessSocket\
            or cfg.concurrency == 'thread':
            cfg.set('workers', 0)
        # Open the socket and bind to address
        address = self.cfg.address
        if address:
            socket = pulsar.create_socket(address, backlog=self.cfg.backlog)
        else:
            raise pulsar.ImproperlyConfigured('Could not open a socket. '
                                              'No address to bind to')
        self.logger.info('Listening on %s', socket)
        monitor.params.socket = socket
        self.address = socket.name
    
    def worker_start(self, worker):
        # Start the worker by starting the socket server
        s = self.socket_server_factory(worker, worker.params.socket,
                                       timeout=self.cfg.keepalive)
        # We add the file descriptor handler
        s.on_connection_callbacks.append(worker.handle_fd_event)
        worker.socket_server = s
    
    def worker_stop(self, worker):
        if hasattr(worker, 'socket_server'):
            # we don't shut down the socket, simply remove all active
            # connections and othe clean up operations.
            worker.socket_server.on_close()
        
    def on_event(self, worker, fd, events):
        connection = worker.socket_server.accept()
        if connection is not None:
            return connection.on_closed
    
    def on_info(self, worker, data):
        server = worker.socket_server
        data['socket'] = {'listen_on': server.address,
                          'read_timeout': server.timeout,
                          'active_connections': server.active_connections,
                          'received_connections': server.received}
        return data



import os
import re
import sys
import unittest
import logging

from pulsar.utils.importer import import_module

default_logger = logging.getLogger('pulsar.apps.test')


__all__ = ['TestLoader']


def issubclass_safe(cls, base_cls):
    try:
        return issubclass(cls, base_cls)
    except TypeError:
        return False

class TestLoader(object):
    '''Classes used by the :class:`TestSuite` to aggregate tests
from a list of paths. The way it works is simple,
you give a *root* directory and a list of submodules where to look for tests.

:parameter root: root path passed by the :class:`TestSuite`. 
:parameter modules: list (or tuple) of entries where to look for tests.
    An entry can be a string indicating the **dotted** path relative to the
    **root** directory or a two element tuple with the same dotted path as
    first argument and a **pattern** which files must match in order to be
    included in the search. A third element in the tuple can be used to specify
    the top level tag for all tests in this entry.
    For example::

        modules = ['test', ('bla', None, 'foo'), ('examples','test')]

    loads
    
     * all tests from modules in the ``test`` directory,
     * all tests from the ``bla`` directory with top level tag ``foo``,
     * all tests from the ``examples`` directory matching the test pattern.
     
    All top level modules will be added to the python ``path``.

:parameter runner: Instance of the test runner.

The are very :ref:`simple rules <apps-test-loading>` followed for
importing tests.
'''
    def __init__(self, root, modules, runner, logger=None):
        self.runner = runner
        self.logger = logger or default_logger
        self.root = root
        self.modules = []
        for mod in modules:
            if isinstance(mod, str):
                mod = (mod, None, None)
            if len(mod) < 3:
                mod = tuple(mod) + (None,) * (3 - len(mod))
            self.modules.append(mod)

    def __repr__(self):
        return self.root
    __str__ = __repr__

    def alltags(self, tag):
        bits = tag.split('.')
        tag,rest = bits[0],bits[1:]
        yield tag
        for b in rest:
            tag += '.' + b
            yield tag

    def checktag(self, tag, import_tags):
        '''return ``True`` if ``tag`` is in ``import_tags``.'''
        if import_tags:
            c = 0
            alltags = list(self.alltags(tag))
            for import_tag in import_tags:
                allitags = list(self.alltags(import_tag))
                for bit in alltags:
                    if bit == import_tag:
                        return 2
                    elif bit in allitags:
                        c = 1
            return c
        else:
            return 2

    def testclasses(self, tags=None):
        pt = ', '.join(tags) if tags else 'all'
        self.logger.info('Load test classes for %s tags', pt)
        for tag, mod in self.testmodules(tags):
            if tags:
                skip = True
                for bit in self.alltags(tag):
                    if bit in tags:
                        skip = False
                        break
                if skip:
                    continue
            for name in dir(mod):
                obj = getattr(mod, name)
                if issubclass_safe(obj, unittest.TestCase):
                    yield tag, obj
    
    def testmodules(self, tags=None):
        '''Generator of tag, test modules pairs.'''
        for name, pattern, tag in self.modules:
            if pattern and '*' in pattern:
                pattern = re.compile(pattern.replace('*', '(.*)'))
            names = name.split('.') if name else ()
            absolute_path = os.path.join(self.root, *names)
            if os.path.isdir(absolute_path):
                pathbase = os.path.dirname(absolute_path)
                if pathbase not in sys.path:
                    sys.path.append(pathbase)
                name = names[-1]
                stags = (tag,) if tag else ()
                for tag, mod in self.get_tests(absolute_path, name, pattern,
                                               import_tags=tags, tags=stags):
                    yield tag, mod
            elif os.path.isfile(absolute_path + '.py'):
                tag = tag or name
                mod = self.import_module(name)
                if mod:
                    yield tag, mod
            else:
                raise ValueError('%s cannot be found in %s directory.'\
                                  % (name, self.root))

    def get_tests(self, path, dotted_path, pattern, import_tags=None,
                  tags=(), parent=None):
        '''Collect python modules for testing and return a generator of
tag,module pairs.

:parameter path: directory path where to search. Files starting with ``_``
    or ``.`` are excluded from the search, as well as non-python files.

:parameter dotted_path: the dotted python path equivalent of ``path``.

:parameter parent: the parent module for the current one. This parameter
    is passed by this function recursively.'''
        for mod_name in os.listdir(path):
            if mod_name.startswith('_') or mod_name.startswith('.'):
                continue
            mod_path = os.path.join(path, mod_name)
            if os.path.isfile(mod_path):
                if mod_name.endswith('.py'):
                    mod_name = mod_name.split('.')[0]
                else:
                    continue
            addtag = mod_name
            npattern = pattern
            if pattern:
                if hasattr(pattern, 'search'):
                    p = pattern.search(mod_name)
                    if p:
                        npattern = None
                        addtag = p.groups(0)
                    else:
                        addtag = False
                elif pattern == mod_name:
                    addtag = False
                    npattern = None
                if npattern and os.path.isfile(mod_path):
                    # skip the import
                    continue
            if dotted_path:
                mod_dotted_path = '%s.%s' % (dotted_path, mod_name)
            else:
                tags = (mod_name,)
                mod_dotted_path = mod_name
            module = self.import_module(mod_dotted_path, mod_path, parent)
            if not module:
                continue
            ctags = tags + (addtag,) if addtag else tags
            tag = '.'.join(ctags)
            c = self.checktag(tag, import_tags)
            if not c:
                continue
            counter = 0
            if os.path.isdir(mod_path):
                counter = 0
                # Recursively import modules
                for tag, mod in self.get_tests(mod_path,
                                               mod_dotted_path,
                                               npattern,
                                               import_tags,
                                               ctags,
                                               parent = module):
                    counter += 1
                    yield tag, mod
            # No submodules
            if not counter and c == 2:
                yield tag, module

    def import_module(self, name, path=None, parent=None):
        imp = True
        if path and os.path.isdir(path):
            imp = False
            # import only if it has a __init__.py file
            for sname in os.listdir(path):
                if sname == '__init__.py':
                    imp = True
                    break
        if imp:
            try:
                mod = import_module(name)
                if getattr(mod, '__test__', True):
                    return self.runner.import_module(mod, parent)
            except ImportError:
               self.logger.error('failed to import module %s. Skipping.',
                                 name, exc_info=True)
            except:
               self.logger.critical('Failed to import module %s. Skipping.',
                                    name, exc_info=True)




'''A a JSON-RPC Server with a Task Queue for processing tasks.
To run the server type::

    python manage.py
    
Open a new shell and launch python and type::

    >>> from pulsar.apps import rpc
    >>> p = rpc.JsonProxy('http://localhost:8060')
    >>> p.ping()
    'pong'
    >>> p.calc.add(3,4)
    7.0
    >>>
    
'''
try:
    import pulsar
except ImportError:
    import sys
    sys.path.append('../../')
    import pulsar
from pulsar.apps import rpc, tasks, wsgi

TASK_PATHS = ['sampletasks.*']


class RpcRoot(rpc.PulsarServerCommands,
              tasks.TaskQueueRpcMixin):
    '''The rpc handler which communicates with the task queue'''
    
    def rpc_runpycode(self, request, code=None, **params):
        return self.task_run(request, 'runpycode', code=code, **params)
        
def dummy():
    pass

class server(pulsar.MultiApp):
    
    def __call__(self, actor=None):
        name = self.name
        params = self.params
        tq = tasks.TaskQueue(name=name, tasks_path=TASK_PATHS,
                             script=__file__, **params)
        self.apps.append(tq)
        rpcs = wsgi.WSGIServer(rpc.RpcMiddleware(RpcRoot(tq)),
                               name = '{0}_rpc'.format(tq.name),
                               **params)
        self.apps.append(rpcs)
        return rpcs(actor)
    

if __name__ == '__main__':
    server = server()
    server()
    server.start()




from .access import *
from .defer import *
from .iostream import *
from .mailbox import *
from .eventloop import *
from .proxy import *
from .actor import *
from .arbiter import *
from .monitor import *
from .concurrency import *
from .http import *
from . import commands


'''\
A very Simple Web-Socket example.
To run the server type::

    python manage.py
    
and open a web browser at http://localhost:8060    
'''
import os
import sys
import json
from random import random
import time
try:
    import pulsar
except ImportError: #pragma nocover
    sys.path.append('../../')
    import pulsar
from pulsar.apps import ws, wsgi
from pulsar.utils.httpurl import range


class handle(ws.WS):
    
    def match(self, environ):
        return environ.get('PATH_INFO') in ('/echo', '/data')
    
    def on_message(self, environ, msg):
        path = environ.get('PATH_INFO')
        if path == '/echo':
            return msg
        elif path == '/data':
            return json.dumps([(i,random()) for i in range(100)])


def page(environ, start_response):
    """ This resolves to the web page or the websocket depending on the path."""
    path = environ.get('PATH_INFO')
    if not path or path == '/':
        data = open(os.path.join(os.path.dirname(__file__), 
                     'websocket.html')).read()
        data = data % environ
        start_response('200 OK', [('Content-Type', 'text/html'),
                                  ('Content-Length', str(len(data)))])
        return [pulsar.to_bytes(data)]



def server(**kwargs):
    app = wsgi.WsgiHandler(middleware=(page, ws.WebSocket(handle())))
    return wsgi.WSGIServer(callable=app, **kwargs)


if __name__ == '__main__':  #pragma nocover
    server().start()



import os
import sys
from time import time

import pulsar
from pulsar.utils.structures import AttributeDictionary

from . import proxy
from .actor import Actor, ACTOR_STATES, ACTOR_TERMINATE_TIMEOUT,\
                     ACTOR_STOPPING_LOOPS
from .eventloop import setid
from .concurrency import concurrency
from .defer import async, iteritems, itervalues, range, NOT_DONE
from .mailbox import Queue, mailbox


__all__ = ['Monitor', 'PoolMixin']


def _spawn_actor(cls, commands_set=None, monitor=None, cfg=None, name=None,
                 aid=None, **kw):
    # Internal function which spawns a new Actor and return its
    # ActorProxyMonitor.
    # *cls* is the Actor class
    # *monitor* can be either the ariber or a monitor
    kind = None
    if issubclass(cls, PoolMixin):
        kind = 'monitor'
    if cfg is None:
        if monitor:
            cfg = monitor.cfg.copy()
        else:
            cfg = pulsar.Config()
    if monitor:
        params = monitor.actorparams()
        name = params.pop('name', name)
        aid = params.pop('aid', aid)
        commands_set = params.pop('commands_set', commands_set)
    else:
        if kind != 'monitor':
            raise TypeError('class %s not a valid monitor' % cls)
        params = {}
    commands_set = set(commands_set or proxy.actor_commands)
    for key, value in iteritems(kw):
        if key in cfg.settings:
            cfg.set(key, value)
        else:
            params[key] = value
    #
    if monitor:
        if not kind:
            if not issubclass(cls, Actor):
                raise TypeError('Class %s not a valid actor.' % cls)
            kind = cfg.concurrency
    if not kind:
        raise TypeError('Cannot spawn class %s. not a valid concurrency.' % cls)
    actor_proxy = concurrency(kind, cls, monitor, commands_set, cfg,
                              name=name, aid=aid, **params)
    # Add to the list of managed actors if this is a remote actor
    if isinstance(actor_proxy, Actor):
        return actor_proxy
    else:
        actor_proxy.monitor = monitor
        monitor.spawning_actors[actor_proxy.aid] = actor_proxy
        deferred = proxy.ActorProxyDeferred(actor_proxy)
        actor_proxy.start()
        return deferred


class PoolMixin(object):
    '''Not an actor per se, this is a mixin for :class:`Actor`
which manages a pool (group) of actors. Given an :attr:`actor_class`
it makes sure there are always :attr:`cfg.workers` alive.
It is used by both the :class:`Arbiter` and the :class:`Monitor` classes.

.. attribute:: managed_actors

    dictionary with keys given by actor's ids and values by
    :class:`ActorProxyMonitor` instances. These are the actors managed by the
    pool.
    
.. attribute:: spawning_actors

    A dictionary of :class:`ActorProxyMonitor` which are in the process of
    being spawned.

'''
    CLOSE_TIMEOUT = 30000000000000
    actor_class = Actor
    '''The class derived form :class:`Actor` which the monitor manages
during its life time.

    Default: :class:`Actor`'''

    def on_start(self):
        self.spawning_actors = {}
        self.managed_actors = {}
        self.actor_class = self.params.pop('actor_class') or self.actor_class

    def active(self):
        return self.running()

    def spawn(self, actor_class=None, linked_actors=None, montitor=None,
              **params):
        '''Spawn a new :class:`Actor` and return its
:class:`ActorProxyMonitor`.'''
        actor_class = actor_class or self.actor_class
        if linked_actors:
            params['linked_actors'] =\
                dict(((aid, p.proxy) for aid, p in iteritems(linked_actors)))
        return _spawn_actor(actor_class, monitor=self, **params)

    def actorparams(self):
        '''Return a dictionary of parameters to be passed to the
spawn method when creating new actors.'''
        arbiter = self.arbiter or self
        params = dict(self.params)
        params['monitors'] = arbiter.get_all_monitors()
        return params

    def get_actor(self, aid):
        a = Actor.get_actor(self, aid)
        if not a:
            a = self.spawning_actors.get(aid)
        return a

    def manage_actors(self, terminate=False, stop=False, manage=True):
        '''Remove :class:`Actor` which are not alive from the
:class:`PoolMixin.managed_actors` and return the number of actors still alive.

:parameter terminate: if ``True`` force termination of alive actors.
:parameter stop: if ``True`` stops all alive actor.
:parameter manage: if ``True`` it checks if alive actors are still responsive.
'''
        MANAGED = self.managed_actors
        SPAWNING = self.spawning_actors
        LINKED = self.linked_actors
        alive = 0
        ACTORS = list(iteritems(MANAGED))
        # WHEN TERMINATING OR STOPPING WE INCLUDE THE ACTORS WHICH ARE SPAWNING
        shutting_down = terminate or stop
        if shutting_down:
            ACTORS.extend(iteritems(SPAWNING))
        # Loop over MANAGED ACTORS PROXY MONITORS
        for aid, actor in ACTORS:
            if not actor.is_alive():
                actor.join(ACTOR_TERMINATE_TIMEOUT)
                MANAGED.pop(aid, None)
                SPAWNING.pop(aid, None)
                LINKED.pop(aid, None)
            else:
                alive += 1
                if shutting_down:
                    # terminate if there is not mailbox (ther actor has
                    # registered with its monitor yet).
                    if terminate or actor.mailbox is None:
                        actor.terminate()
                        actor.join(ACTOR_TERMINATE_TIMEOUT)
                    else:
                        actor.stop(self)
                elif manage:
                    self.manage_actor(actor)
        return alive

    def manage_actor(self, actor):
        '''If an actor failed to notify itself to the arbiter for more than
the timeout. Stop the arbiter.'''
        stopping_loops = actor.stopping_loops
        if self.running() and (stopping_loops or actor.notified):
            gap = time() - actor.notified
            if gap > actor.cfg.timeout or stopping_loops:
                if stopping_loops < ACTOR_STOPPING_LOOPS:
                    if not stopping_loops:
                        self.logger.info('Stopping %s. Timeout.', actor)
                        self.send(actor, 'stop')
                    actor.stopping_loops += 1
                else:
                    self.logger.warn('Terminating %s. Timeout.', actor)
                    actor.terminate()
                    actor.join(ACTOR_TERMINATE_TIMEOUT)

    def spawn_actors(self):
        '''Spawn new actors if needed. If the :class:`PoolMixin` is spawning
do nothing.'''
        to_spawn = self.cfg.workers - len(self.managed_actors)
        if self.cfg.workers and to_spawn > 0 and not self.spawning_actors:
            for _ in range(to_spawn):
                self.spawn()

    def stop_actors(self):
        """Maintain the number of workers by spawning or killing
as required."""
        if self.cfg.workers:
            num_to_kill = len(self.managed_actors) - self.cfg.workers
            for i in range(num_to_kill, 0, -1):
                w, kage = 0, sys.maxsize
                for worker in itervalues(self.managed_actors):
                    age = worker.impl.age
                    if age < kage:
                        w, kage = w, age
                self.stop_actor(w)

    def stop_actor(self, actor):
        raise NotImplementedError()

    def link_actor(self, proxy, address):
        # Override the link_actor from Actor class
        self.logger.debug('Registering %s address %s', proxy, address)
        if proxy.aid not in self.spawning_actors:
            raise RuntimeError('Could not retrieve proxy %s.' % proxy)
        proxy_monitor = self.spawning_actors.pop(proxy.aid)
        proxy_monitor.info['last_notified'] = time()
        on_address = proxy_monitor.on_address
        delattr(proxy_monitor,'on_address')
        try:
            if not address:
                raise ValueError('No address received')
        except:
            on_address.callback(sys.exc_info())
        else:
            proxy_monitor.address = address
            self.managed_actors[proxy.aid] = proxy_monitor
            self.linked_actors[proxy.aid] = proxy
            if self.arbiter:
                # link also the arbiter
                self.arbiter.linked_actors[proxy.aid] = proxy
            on_address.callback(proxy_monitor.proxy)
        return proxy_monitor.proxy

    @async()
    def close_actors(self):
        '''Close all managed :class:`Actor`.'''
        start = time()
        # Stop all of them
        to_stop = self.manage_actors(stop=True)
        while to_stop:
            yield NOT_DONE
            to_stop = self.manage_actors(manage=False)
            dt = time() - start
            if dt > self.CLOSE_TIMEOUT:
                self.logger.warn('Cannot stop %s actors.', to_stop)
                to_stop = self.manage_actors(terminate=True)
                self.logger.warn('terminated %s actors.', to_stop)
                to_stop = 0


class Monitor(PoolMixin, Actor):
    '''A monitor is a **very** special :class:`Actor` and :class:`PoolMixin`
which shares the same :class:`IOLoop` with the :class:`Arbiter` and
therefore lives in the main process domain.
The Arbiter manages monitors which in turn manage a set of :class:`Actor`
performing similar tasks.

In other words, you may have a monitor managing actors for serving HTTP
requests on a given port, another monitor managing actors consuming tasks
from a task queue and so forth. You can think of :class:`Monitor` as
managers of pools of :class:`Actor`.

Monitors are created by invoking the :meth:`Arbiter.add_monitor`
functions and not by directly invoking the constructor. Therefore
adding a new monitor to the arbiter follows the pattern::

    import pulsar

    m = pulsar.arbiter().add_monitor(pulsar.Monitor,'mymonitor')

You can also create a monitor with a distributed queue as IO mechanism::

    from multiprocessing import Queue
    import pulsar

    m = pulsar.arbiter().add_monitor(pulsar.Monitor,
                                     'mymonitor',
                                     ioqueue = Queue())

Monitors with distributed queues manage :ref:`CPU-bound actors <cpubound>`.
'''
    socket = None

    @property
    def cpubound(self):
        return False

    def is_process(self):
        return False

    def is_monitor(self):
        return True

    def monitor_task(self):
        '''Monitor specific task called by the :meth:`Monitor.periodic_task`.
By default it does nothing. Override if you need to.'''
        pass

    def periodic_task(self):
        '''Overrides the :meth:`Actor.on_task`
:ref:`actor callback <actor-callbacks>` to perform
the monitor :class:`IOLoop` tasks, which are:

* To maintain a responsive set of actors ready to perform their duty.
* To perform its own tasks.

The implementation goes as following:

* It calls :meth:`PoolMixin.manage_actors` which removes from the live
  actors dictionary all actors which are not alive.
* Spawn new actors if required by calling :meth:`PoolMixin.spawn_actors`
  and :meth:`PoolMixin.stop_actors`.
* Call :meth:`Monitor.monitor_task` which performs the monitor specific
  task.

Users shouldn't need to override this method, but use
:meth:`Monitor.monitor_task` instead.'''
        if self.running():
            self.manage_actors()
            self.spawn_actors()
            self.stop_actors()
            self.monitor_task()
        self.ioloop.add_callback(self.periodic_task, False)

    # HOOKS
    def on_stop(self):
        return self.close_actors()

    @property
    def multithread(self):
        return self.cfg.concurrency == 'thread'

    @property
    def multiprocess(self):
        return self.cfg.concurrency == 'process'

    def actorparams(self):
        '''Spawn a new actor and add its :class:`ActorProxyMonitor`
 to the :attr:`PoolMixin.managed_actors` dictionary.'''
        p = super(Monitor, self).actorparams()
        p.update({'ioqueue': self.ioqueue,
                  'commands_set': self.impl.commands_set})
        return p

    def stop_actor(self, actor):
        if not actor.is_alive():
            self.managed_actors.pop(actor.aid)
        else:
            return actor.proxy.stop()

    def info(self):
        data = {'actor': {'actor_class':self.actor_class.__name__,
                          'concurrency':self.cfg.concurrency,
                          'name':self.name,
                          'age':self.impl.age,
                          'workers': len(self.managed_actors)}}
        if not self.started():
            return data
        data['workers'] = [a.info for a in itervalues(self.managed_actors)]
        tq = self.ioqueue
        if tq is not None:
            if isinstance(tq, Queue):
                tqs = 'multiprocessing.Queue'
            else:
                tqs = str(tq)
            try:
                size = tq.qsize()
            except NotImplementedError: #pragma    nocover
                size = 0
            data['queue'] = {'ioqueue': tqs,
                             'ioqueue_size': size}
        return self.on_info(data)

    def proxy_mailbox(address):
        return self.arbiter.proxy_mailboxes.get(address)

    def get_actor(self, aid):
        #Delegate get_actor to the arbiter
        a = super(Monitor, self).get_actor(aid)
        if a is None:
            a = self.arbiter.get_actor(aid)
        return a

    # OVERRIDES INTERNALS
    def _setup_ioloop(self):
        self.requestloop = self.arbiter.requestloop
        self.mailbox = mailbox(self)
        self.monitors = self.arbiter.monitors
        
    def _run(self):
        pass



from time import time

from pulsar import AuthenticationError
from . import proxy 

#############################################################  COMMANDS

@proxy.command()
def auth(client, actor, password):
    '''Request for authentication in a password protected Pulsar server.
Pulsar can be instructed to require a password before allowing clients to
execute commands.'''
    p = actor.cfg.get('password')
    if p and p != password:
        client.connection.authenticated = False
        raise AuthenticationError()
    else:
        client.connection.authenticated = True
        return True

@proxy.command()
def ping(client, actor):
    return 'pong'

@proxy.command()
def echo(client, actor, message):
    '''Returns mmessage'''
    return message

@proxy.command(ack=False)
def quit(client, actor):
    '''Ask the server to close the connection. The connection is closed as
soon as all pending replies have been written to the client.'''
    client.connection.close()

@proxy.command(authenticated=True)
def info(client, actor):
    ''' Returns information and statistics about the server as a json string'''
    return actor.info()

@proxy.command(authenticated=True)
def shutdown(client, actor):
    '''Shutdown the server'''
    return actor.stop()

@proxy.command(authenticated=True)
def config(client, actor, setget, name, *value):
    setget = setget.lower()
    if setget == 'get':
        if len(values) > 0:
            raise CommandError('"config get" accept only one parameter')
        return client.actor.cfg.get(name)
    elif setget == 'set':
        if len(values) > 1:
            raise CommandError()
        client.actor.cfg.set(name, value[0])
    else:
        raise CommandError('config must be followed by set or get')


###########################################################  INTERNAL COMMANDS
@proxy.command(internal=True)
def mailbox_address(client, actor, caller, address):
    '''The remote *caller* register its mailbox ``address``.'''
    actor.link_actor(caller, address)
    return actor.proxy

@proxy.command(internal=True)
def run(client, actor, caller, callable, *args, **kwargs):
    '''Execute a python script in the server'''
    return callable(actor, *args, **kwargs)

@proxy.command(ack=False, internal=True)
def stop(client, actor, caller):
    '''Stop the actor from running. Same as shut_down but for internal use'''
    return actor.stop()
    
@proxy.command(ack=False, internal=True)
def notify(client, actor, caller, info):
    '''caller notify itself.'''
    if isinstance(info, dict):
        info['last_notified'] = time()
        caller.info = info
    
###########################################################  ARBITER COMMANDS
@proxy.command(internal=True, commands_set=proxy.arbiter_commands)
def spawn(client, actor, caller, linked_actors=None, **kwargs):
    linked_actors = linked_actors if linked_actors is not None else {}
    linked_actors[caller.aid] = caller
    return actor.spawn(linked_actors=linked_actors, **kwargs)
     
@proxy.command(authenticated=True, commands_set=proxy.arbiter_commands)
def kill_actor(client, actor, aid=None):
    '''Kill an actor with id ``aid``'''
    if not aid:
        return actor.stop()
    else:
        a = actor.get_actor(aid)
        if a:
            a.stop()
            return 'stopped %s' % a
        else:
            actor.logger.info('Could not kill "%s" no such actor', aid)
        


import logging

from pulsar import ispy3k

if ispy3k:
    from xmlrpc import client as rpc
else:
    import xmlrpclib as rpc


__all__ = ['INVALID_RPC',
           'METHOD_NOT_FOUND',
           'INVALID_METHOD_PARAMS',
           'INTERNAL_ERROR',
           'REQUIRES_AUTHENTICATION',
           'Fault',
           'InternalError',
           'InvalidRequest',
           'NoSuchFunction',
           'InvalidParams',
           'RequiresAuthentication']

_errclasses = {}
INVALID_RPC = rpc.INVALID_XMLRPC
METHOD_NOT_FOUND = rpc.METHOD_NOT_FOUND
INVALID_METHOD_PARAMS = rpc.INVALID_METHOD_PARAMS
INTERNAL_ERROR = rpc.INTERNAL_ERROR
REQUIRES_AUTHENTICATION =-32000


class FaultType(type):
    '''A metaclass for rpc handlers.
Add a limited ammount of magic to RPC handlers.'''
    def __new__(cls, name, bases, attrs):
        newcls = super(FaultType, cls).__new__(cls, name, bases, attrs)
        code = attrs.get('FAULT_CODE')
        if code:
            _errclasses[code] = newcls
        return newcls


class Fault(FaultType('FaultBase', (rpc.Fault,), {})):
    FAULT_CODE = None
    def __init__(self, *args, **kwargs):
        rpc.Fault.__init__(self, self.FAULT_CODE, *args, **kwargs)

    def __repr__(self):
        return '%s %s: %s' % (self.__class__.__name__, self.faultCode,
                               self.faultString)


class InternalError(Fault):
    FAULT_CODE = INTERNAL_ERROR


class InvalidRequest(Fault):
    FAULT_CODE = INVALID_RPC


class NoSuchFunction(InvalidRequest):
    FAULT_CODE = METHOD_NOT_FOUND


class InvalidParams(InvalidRequest):
    FAULT_CODE = INVALID_METHOD_PARAMS


class RequiresAuthentication(Fault):
    FAULT_CODE = REQUIRES_AUTHENTICATION


def handle(err):
    if isinstance(err, Fault):
        return err
    return InternalError(str(err))


def exception(code, *args, **extra):
    klass = _errclasses.get(code, InternalError)
    return klass(*args, **extra)


class JSONEncodeException(Exception):
    '''JSON RPC encode error'''
    pass


class ConcurrencyError(Exception):
    pass


class TimeLimitExceeded(ConcurrencyError):
    """The time limit has been exceeded and the job has been terminated."""
    pass


class SoftTimeLimitExceeded(ConcurrencyError):
    """The soft time limit has been exceeded. This exception is raised
    to give the task a chance to clean up."""
    pass


class AlreadyStarted(ConcurrencyError):
    '''ConcurrentPool already started'''


class TimeoutError(ConcurrencyError):
    pass


class MaybeEncodingError(Exception):
    """Wraps unpickleable object."""

    def __init__(self, exc, value):
        self.exc = str(exc)
        self.value = repr(value)
        super(MaybeEncodingError, self).__init__(self.exc, self.value)

    def __repr__(self):
        return "<MaybeEncodingError: %s>" % str(self)

    def __str__(self):
        return "Error sending result: '%s'. Reason: '%s'." % (
                    self.value, self.exc)


class SchedulingError(Exception):
    pass
